"""
Set of function related to handling the configuration of this option.
"""
from pkglts.dependency import Dependency


def require(purpose, cfg):
    """List of requirements for this option for a given purpose.

    Args:
        purpose (str): either 'option', 'setup', 'install' or 'dvlpt'
        cfg (Config):  current package configuration

    Returns:
        (list of Dependency)
    """
    del cfg

    if purpose == 'option':
        options = ['pysetup']
        return [Dependency(name) for name in options]

    return []



"""
Set of function to extend jinja2 environment.
"""
from pkglts.local import pkg_full_name
from pkglts.option.doc import fmt_badge


def auto_classifiers(cfg):
    """Generate a list of calssifiers for pypi from all sections of config.

    Args:
        cfg (Config):  current package configuration

    Returns:
        (list of str)
    """
    items = set(cfg['pypi']['classifiers'])

    # add license item
    # TODO

    # add intended versions items
    intended_versions = cfg['pysetup']['intended_versions']
    if intended_versions:
        items.add("Programming Language :: Python")

        ver_cla_tpl = "Programming Language :: Python :: %s.%s"
        major_versions = set()
        for ver in intended_versions:
            items.add(ver_cla_tpl % (ver[0], ver[1]))
            major_versions.add(ver[0])

        ver_cla_tpl = "Programming Language :: Python :: %s"
        for ver in major_versions:
            items.add(ver_cla_tpl % ver)

        if len(major_versions) == 1:
            ver, = major_versions
            items.add("Programming Language :: Python :: %s :: Only" % ver)

    return sorted(items)


def environment_extensions(cfg):
    """Add more functionality to an environment.

    Args:
        cfg (Config):  current package configuration

    Returns:
        dict of str: any
    """
    url = "badge.fury.io/py/%s" % pkg_full_name(cfg)
    img = url + ".svg"
    badge = fmt_badge(img, url, "PyPI version", cfg['doc']['fmt'])

    return {"badge": badge,
            "auto_classifiers": auto_classifiers(cfg)}



from pkglts.config_management import Config


def test_badge():
    cfg = Config(dict(doc={'fmt': 'rst'}, readthedocs={'project': "project"}))
    cfg.load_extra()
    assert ".. image:" in cfg._env.globals['readthedocs'].badge



import pytest

from pkglts.install_env.pip_front_end import install, installed_packages

from .venv_tools import clear_venv, create_venv

__test__ = False


@pytest.fixture()
def venv():
    name = "tyti_pip"
    mem = {}

    create_venv(venv, mem)

    yield name

    clear_venv(venv, mem)


def test_installed_packages():
    pkgs = set(installed_packages())
    assert "wheel" in pkgs


def test_install(venv):
    # create virtualenv
    assert "my-first-p" not in installed_packages()

    # install local wheel
    install("test/test_install_env/my_first_p-0.0.4-py2.py3-none-any.whl")

    # test new package still not in virtualenv
    assert "my-first-p" in installed_packages()



from pkglts.config_management import Config
from pkglts.option.readthedocs.config import check, require, update_parameters


def test_update_parameters():
    cfg = {}
    update_parameters(cfg)
    assert len(cfg['readthedocs']) == 1


def test_config_check_project_exists():
    cfg = Config(dict(readthedocs={'project': ""}))
    assert 'readthedocs.project' in check(cfg)


def test_require():
    cfg = Config(dict(readthedocs={}))

    assert len(require('option', cfg)) == 3
    assert len(require('setup', cfg)) == 0
    assert len(require('install', cfg)) == 0
    assert len(require('dvlpt', cfg)) == 0



from pkglts.config_management import Config
from pkglts.option.data.config import require, update_parameters


def test_update_parameters():
    cfg = {}
    update_parameters(cfg)
    assert len(cfg['data']) == 2


def test_require():
    cfg = Config(dict(base={}, data={}))
    
    assert len(require('option', cfg)) == 1
    assert len(require('setup', cfg)) == 0
    assert len(require('install', cfg)) == 0
    assert len(require('dvlpt', cfg)) == 0



"""
Set of function to extend jinja2 environment.
"""
from pkglts.option.doc import fmt_badge


def environment_extensions(cfg):
    """Add more functionality to an environment.

    Args:
        cfg (Config):  current package configuration

    Returns:
        dict of str: any
    """
    owner = cfg['github']['owner']
    project = cfg['github']['project'].replace("_", "-")
    token = cfg['appveyor']['token']

    url = "ci.appveyor.com/project/%s/%s/branch/master" % (owner, project)
    img = "ci.appveyor.com/api/projects/status/%s/branch/master?svg=true" % token
    badge = fmt_badge(img, url, "Appveyor build status", cfg['doc']['fmt'])

    return {"badge": badge}



from pkglts.config_management import Config
from pkglts.option.test.config import check, require, update_parameters


def test_update_parameters():
    cfg = {}
    update_parameters(cfg)
    assert len(cfg['test']) == 1


def test_config_check_suite_names():
    for name in ('walou', ' nose'):
        cfg = Config(dict(test={'suite_name': name}))
        assert 'test.suite_name' in check(cfg)


def test_require():
    for suite_name in ('pytest', 'nose'):
        cfg = Config(dict(base={}, test={'suite_name': suite_name}))

        assert len(require('option', cfg)) == 1
        assert len(require('setup', cfg)) == 0
        assert len(require('install', cfg)) == 0
        assert len(require('dvlpt', cfg)) == 2



"""
Set of function related to handling the configuration of this option.
"""
from os.path import abspath, basename


def update_parameters(cfg):
    """Update config with parameters necessary for this option.

    Notes: create a section with option name to store params.

    Args:
        cfg (dict): dict of option parameters as seen in pkg_cfg.json

    Returns:
        None: update in place
    """
    sec = dict(
        pkgname=basename(abspath(".")),
        namespace=None,
        namespace_method="pkg_util",
        url=None,
        authors=[("moi", "moi@email.com")]
    )
    cfg['base'] = sec


def is_valid_identifier(name):
    """ Check that name is a valid python identifier
    sort of back port of "".isidentifier()
    """
    try:
        compile("%s=1" % name, "test", 'single')
        return True
    except SyntaxError:
        return False


def check(cfg):
    """Check the validity of parameters in working environment.

    Args:
        cfg (Config):  current package configuration

    Returns:
        (list of str): list of faulty parameters
    """
    invalids = []
    pkgname = cfg['base']['pkgname']
    namespace = cfg['base']['namespace']

    if "." in pkgname:
        invalids.append('base.pkgname')
    elif not is_valid_identifier(pkgname):
        invalids.append('base.pkgname')

    if namespace is not None:
        if "." in namespace:
            invalids.append('base.namespace')
        elif not is_valid_identifier(namespace):
            invalids.append('base.namespace')

    if cfg['base']['namespace_method'] not in ("pkg_util", "setuptools", "P3.3>"):
        invalids.append("base.namespace_method")

    return invalids


def require(purpose, cfg):
    """List of requirements for this option for a given purpose.

    Args:
        purpose (str): either 'option', 'setup', 'install' or 'dvlpt'
        cfg (Config):  current package configuration

    Returns:
        (list of Dependency)
    """
    del cfg, purpose

    return []



from pkglts.config_management import Config, DEFAULT_CFG
from pkglts.manage_tools import check_option_parameters


def test_check_handle_option_config_not_existing():
    cfg = Config(DEFAULT_CFG)
    assert len(check_option_parameters("toto", cfg)) == 0


def test_check_handle_option_defines_no_check():
    cfg = Config(DEFAULT_CFG)
    assert len(check_option_parameters("coverage", cfg)) == 0



"""
Set of function to extend jinja2 environment.
"""
from pkglts.option.doc import fmt_badge


def environment_extensions(cfg):
    """Add more functionality to an environment.

    Args:
        cfg (Config):  current package configuration

    Returns:
        dict of str: any
    """
    project = cfg['readthedocs']['project']
    project = project.replace(".", "")
    url = "%s.readthedocs.io/en/latest/?badge=latest" % project
    img = "readthedocs.org/projects/%s/badge/?version=latest" % project
    badge = fmt_badge(img, url, "Documentation status", cfg['doc']['fmt'])

    return {"badge": badge}



from pkglts.config_management import Config
from pkglts.option.coveralls.config import require


def test_require():
    cfg = Config(dict(coveralls={}))

    assert len(require('option', cfg)) == 2
    assert len(require('setup', cfg)) == 0
    assert len(require('install', cfg)) == 0
    assert len(require('dvlpt', cfg)) == 1



from copy import deepcopy

from pkglts.config_management import Config


def test_pkg_url_empty_default():
    cfg = Config(dict(pysetup={"intended_versions": ["27"],
                               'require': []}))
    cfg.load_extra()
    assert cfg._env.globals['pysetup'].pkg_url == ""


def test_pkg_url_look_multiple_places():
    tpl_cfg = dict(base={'pkgname': 'toto', 'namespace': 'oa', 'url': None},
                   doc={'fmt': 'rst'},
                   github={'url': None, 'project': 'project', 'owner': 'toto'},
                   gitlab={'url': None, 'project': 'project', 'owner': 'toto'},
                   pypi={'classifiers': [], 'url': None},
                   readthedocs={'project': 'project'},
                   pysetup={'intended_versions': ["27"],
                            'require': []})
    
    for name in ("base", "github", "gitlab", "pypi", "readthedocs"):
        loc_cfg = deepcopy(tpl_cfg)
        loc_cfg[name]['url'] = name
        cfg = Config(loc_cfg)
        cfg.load_extra()
        assert cfg._env.globals['pysetup'].pkg_url == name


def test_requirements():
    cfg = Config(dict(test={'suite_name': 'pytest'},
                      pysetup={'intended_versions': ["27"],
                               'require': []}))
    cfg.load_extra()
    assert len(cfg._env.globals['pysetup'].requirements('install')) == 0
    assert len(cfg._env.globals['pysetup'].requirements('dvlpt')) == 2



from glob import glob
from os import path

import pkglts
from pkglts.config_management import Config
from pkglts.option_tools import available_options


def test_options_expose_parameters():
    # walk through all possible options defined by pkglts
    option_basedir = path.join(path.dirname(pkglts.__file__), 'option')
    for pth in glob("{}/*/".format(option_basedir)):
        option_name = path.basename(path.dirname(pth))
        if not option_name.startswith("_"):
            # check 'update_parameters' exists for each option
            try:
                opt = available_options[option_name]
                cfg = {}
                opt.update_parameters(cfg)
                assert len(cfg) == 1
            except KeyError:
                assert False


def test_require_correctly_defined():
    cfg = Config(dict(base={}, doc={'fmt': 'rst'}, sphinx={'theme': 'default'}, test={'suite_name': 'pytest'}))

    # walk through all possible options
    option_basedir = path.join(path.dirname(pkglts.__file__), 'option')
    for pth in glob("{}/*/".format(option_basedir)):
        option_name = path.basename(path.dirname(pth))
        if not option_name.startswith("_"):
            # check 'require' function exists for each option
            try:
                opt = available_options[option_name]
                assert len(opt.require('option', cfg)) >= 0
                assert len(opt.require('setup', cfg)) >= 0
                assert len(opt.require('install', cfg)) >= 0
                assert len(opt.require('dvlpt', cfg)) >= 0
            except ImportError:
                assert False



from pkglts.config_management import Config
from pkglts.option.version.config import check, require, update_parameters


def test_update_parameters():
    cfg = {}
    update_parameters(cfg)
    assert len(cfg['version']) == 3


def test_config_check_version_numbers_are_valid():
    cfg = Config(dict(version={'major': "", 'minor': "", 'post': ""}))
    assert 'version.major' in check(cfg)
    assert 'version.minor' in check(cfg)
    assert 'version.post' in check(cfg)
    cfg = Config(dict(version={'major': "a", 'minor': "a", 'post': "a"}))
    assert 'version.major' in check(cfg)
    assert 'version.minor' in check(cfg)
    assert 'version.post' in check(cfg)
    cfg = Config(dict(version={'major': "1", 'minor': "1", 'post': "1"}))
    assert 'version.major' in check(cfg)
    assert 'version.minor' in check(cfg)
    assert 'version.post' in check(cfg)
    cfg = Config(dict(version={'major': 1, 'minor': 0, 'post': "2.dev"}))
    assert 'version.post' in check(cfg)


def test_require():
    cfg = Config(dict(base={}, version={}))

    assert len(require('option', cfg)) == 1
    assert len(require('setup', cfg)) == 0
    assert len(require('install', cfg)) == 0
    assert len(require('dvlpt', cfg)) == 0



"""Regroup set of functions that make use of local environment
inside a package. Just a way to normalize pre defined paths.
"""
from os import path

from .templating import render


def pkg_full_name(cfg):
    """Compute name of src dir according to pkgname
    and namespace in environment

    Args:
        cfg (Config):  current package configuration

    Returns:
        (str)
    """
    namespace = cfg['base']['namespace']
    if namespace is None:
        return cfg['base']['pkgname']

    return namespace + "." + cfg['base']['pkgname']


def src_dir(cfg):
    """Compute name of src dir according to pkgname
    and namespace in environment

    Args:
        cfg (Config):  current package configuration

    Returns:
        (str)
    """
    rep = "src"
    namespace = cfg['base']['namespace']
    if namespace is not None:
        rep = rep + "/" + namespace

    pkgname = cfg['base']['pkgname']
    rep = rep + "/" + pkgname

    return rep


def init_namespace_dir(pth, cfg):
    """Populate a directory with specific __init__.py
    for namespace packages.

    Args:
        pth (str): path in which to create the files
        cfg (Config):  current package configuration

    Returns:
        None
    """
    src_pth = path.dirname(__file__) + "/resource/namespace__init__.py.tpl"
    tgt_pth = pth + "/__init__.py"
    render(cfg, src_pth, tgt_pth)



"""
Set of function to extend jinja2 environment.
"""
import logging
from os.path import dirname
from os.path import join as pj

LOGGER = logging.getLogger(__name__)

TPL_DIR = pj(dirname(__file__), "templates")


def get_tpl_path(name):
    """Return a path for license template file.

    Warnings: Do not test if path exists

    Args:
        name (str): name of license to fetch

    Returns:
        (str)
    """
    return pj(TPL_DIR, "%s.txt" % name.lower())


def full_text(cfg):
    """Generate a license
    """
    name = cfg['license']['name']

    # open template
    try:
        with open(get_tpl_path(name), 'r') as fhr:
            cnt = fhr.read()

        return cfg.render(cnt)
    except IOError as err:
        LOGGER.error("unable to find template for given license")
        raise err


def environment_extensions(cfg):
    """Add more functionality to an environment.

    Args:
        cfg (Config):  current package configuration

    Returns:
        dict of str: any
    """
    return {"full_text": full_text(cfg)}



"""
Set of function related to handling the configuration of this option.
"""
from pkglts.dependency import Dependency


def update_parameters(cfg):
    """Update config with parameters necessary for this option.

    Notes: create a section with option name to store params.

    Args:
        cfg (dict): dict of option parameters as seen in pkg_cfg.json

    Returns:
        None: update in place
    """
    cfg['plugin_project'] = {
        "plugin_name": "{{ base.pkgname }}"
    }
    # empty project just for construction
    # so does nothing
    # del cfg
    # return


def is_valid_identifier(name):
    """ Check that name is a valid python identifier
    sort of back port of "".isidentifier()
    """
    try:
        compile("%s=1" % name, "test", 'single')
        return True
    except SyntaxError:
        return False


def check(cfg):
    """Check the validity of parameters in working environment.

    Args:
        cfg (Config):  current package configuration

    Returns:
        (list of str): list of faulty parameters
    """
    invalids = []
    plugin_name = cfg['plugin_project']['plugin_name']

    if "." in plugin_name:
        invalids.append('plugin.plugin_name')
    elif not is_valid_identifier(plugin_name):
        invalids.append('plugin.plugin_name')

    return invalids


def require(purpose, cfg):
    """List of requirements for this option for a given purpose.

    Args:
        purpose (str): either 'option', 'setup', 'install' or 'dvlpt'
        cfg (Config):  current package configuration

    Returns:
        (list of Dependency)
    """
    del cfg

    if purpose == 'option':
        options = ['pysetup', 'data', 'github']
        return [Dependency(name) for name in options]

    if purpose == 'install':
        return [Dependency('pkglts', pkg_mng='pip')]

    return []



from pkglts.config_management import Config
from pkglts.option.coverage.config import require


def test_require():
    cfg = Config(dict(coverage={}, test={'suite_name': 'pytest'}))

    assert len(require('option', cfg)) == 1
    assert len(require('setup', cfg)) == 0
    assert len(require('install', cfg)) == 0
    assert len(require('dvlpt', cfg)) == 2

    cfg = Config(dict(coverage={}, test={'suite_name': 'nose'}))
    assert len(require('dvlpt', cfg)) == 1



from pkglts.config_management import Config
from pkglts.option.sphinx.config import check, require, update_parameters


def test_update_parameters():
    cfg = {}
    update_parameters(cfg)
    assert len(cfg['sphinx']) == 3


def test_config_check_sphinx_theme():
    for theme in (1, None,):
        cfg = Config(dict(doc={'fmt': 'rst'}, sphinx={'theme': theme}))
        assert 'sphinx.theme' in check(cfg)


def test_require():
    cfg = Config(dict(sphinx={'theme': "default"}))

    assert len(require('option', cfg)) == 3
    assert len(require('setup', cfg)) == 0
    assert len(require('install', cfg)) == 0
    assert len(require('dvlpt', cfg)) == 1



from pkglts.config_management import Config
from pkglts.option.appveyor.config import check, require, update_parameters


def test_update_parameters():
    cfg = {}
    update_parameters(cfg)
    assert len(cfg['appveyor']) == 1


def test_config_does_nothing():
    cfg = Config(dict(appveyor={'token': 'toto'}))
    assert len(check(cfg)) == 0


def test_require():
    cfg = Config(dict(base={}, travis={}))
    
    assert len(require('option', cfg)) == 2
    assert len(require('setup', cfg)) == 0
    assert len(require('install', cfg)) == 0
    assert len(require('dvlpt', cfg)) == 0



""" Specific helper function for manage script
"""

import logging
from os import listdir, mkdir
from os.path import basename, exists, isdir, splitext

from .hash_management import compute_hash, pth_as_key
from .install_env.load_front_end import get_install_front_end
from .local import init_namespace_dir
from .option_tools import available_options, get_user_permission
from .templating import render

LOGGER = logging.getLogger(__name__)

TPL_SRC_NAME = "{" + "{ base.pkgname }" + "}"

NON_BIN_EXT = ("", ".bat", ".cfg", ".in", ".ini", ".md", ".no", ".ps1", ".py", ".rst", ".sh",
               ".txt", ".yml", ".yaml")


def ensure_installed_packages(requirements, msg, cfg):
    """Ensure all packages in requirements are installed.

    If not, ask user permission to install them.

    Args:
        requirements (list of Dependency): list of package names to install
                                   if needed
        msg (str): error message to print
        cfg (Config):  current package configuration

    Returns:
        (bool): whether all required packages are installed or not
    """
    ife_name = cfg["_pkglts"]['install_front_end']
    req = [dep.name for dep in requirements
           if dep.package_manager is None or dep.package_manager == ife_name]
    ife = get_install_front_end(ife_name)
    to_install = set(req) - set(ife.installed_packages())
    if to_install:
        print(msg)
        LOGGER.warning("missing packages: " + ", ".join(to_install))
        for name in to_install:
            ife.install(name)
            LOGGER.info("install %s", name)

    return True


def check_option_parameters(name, cfg):
    """Check that the parameters associated to an option are valid.

    Try to import Check function in option dir.

    Args:
        name (str): option name
        cfg (Config):  current package configuration
    """
    try:
        opt = available_options[name]
        return opt.check(cfg)
    except KeyError:
        return []


def update_opt(name, cfg):
    """Update an option of this package.

    Notes: If the option does not exists yet, add it first.
           See the list of available option online

    Args:
        name (str): name of option to add
        cfg (Config):  current package configuration
    """
    LOGGER.info("update option %s", name)

    # test existence of option
    try:
        opt = available_options[name]
    except KeyError:
        raise KeyError("option '%s' does not exists" % name)

    # find other option requirements in repository
    for dep in opt.require('option', cfg):
        option_name = dep.name
        if option_name not in cfg.installed_options():
            print("need to install option '%s' first" % option_name)
            if (cfg["_pkglts"]['auto_install'] or
                    get_user_permission("install")):
                cfg = update_opt(option_name, cfg)
            else:
                return cfg

    # find extra package requirements for setup
    msg = "this option requires some packages to setup"
    if not ensure_installed_packages(opt.require('setup', cfg), msg, cfg):
        print("option installation stopped")
        return cfg

    # find parameters required by option config
    opt.update_parameters(cfg.template())
    cfg.resolve()

    # find extra package requirements for dvlpt
    msg = "this option requires additional packages for developers"
    ensure_installed_packages(opt.require('dvlpt', cfg), msg, cfg)

    return cfg


def _rg_dir(src_pth, tgt_name, tgt_pth, cfg, overwrite_file):
    """Regenerate a directory.

    Notes:
        recursively regenerate all tree rooted on src_pth

    Args:
        src_pth (str): path to directory
        tgt_name (str): name of directory to create
        tgt_pth (str): path to created directory
        cfg (Config):  current package configuration
        overwrite_file (dict): which files to overwrite

    Returns:
        (dict of str, map): hash key of preserved sections
    """
    if tgt_name in ("", "_"):
        return {}

    if not exists(tgt_pth):
        mkdir(tgt_pth)

    return regenerate_dir(src_pth, tgt_pth, cfg, overwrite_file)


def _rg_file(src_pth, tgt_name, tgt_pth, cfg, overwrite_file):
    """Regenerate a single file.

    Args:
        src_pth (str): path to file
        tgt_name (str): name of file to create
        tgt_pth (str): path to created file
        cfg (Config):  current package configuration
        overwrite_file (dict): which files to overwrite

    Returns:
        (dict|None): bid, hash of block content
    """
    if splitext(tgt_name)[0] == "_":
        return None

    LOGGER.debug("render file '%s' into '%s'", src_pth, tgt_pth)

    if overwrite_file.get(pth_as_key(tgt_pth), True):
        _, ext = splitext(tgt_name)
        if ext in NON_BIN_EXT:
            blocks = render(cfg, src_pth, tgt_pth)
            return dict((bid, compute_hash(cnt)) for bid, cnt in blocks)
        else:  # binary file
            if exists(tgt_pth):
                print("overwrite? %s" % tgt_pth)
            else:
                with open(src_pth, 'rb') as fhr:
                    content = fhr.read()
                with open(tgt_pth, 'wb') as fhw:
                    fhw.write(content)


def regenerate_dir(src_dir, tgt_dir, cfg, overwrite_file):
    """Walk all files in src_dir and create/update them on tgt_dir

    Args:
        src_dir (str): path to reference files
        tgt_dir (str): path to target where files will be written
        cfg (Config):  current package configuration
        overwrite_file (dict of str, bool): whether or not to overwrite some
                             files

    Returns:
        (dict of str, map): hash key of preserved sections
    """
    hmap = {}

    for src_name in listdir(src_dir):
        src_pth = src_dir + "/" + src_name
        tgt_name = cfg.render(src_name)
        if tgt_name.endswith(".tpl"):
            tgt_name = tgt_name[:-4]

        tgt_pth = tgt_dir + "/" + tgt_name
        # handle namespace
        if isdir(src_pth) and basename(src_dir) == 'src' and src_name == TPL_SRC_NAME:
            namespace = cfg['base']['namespace']
            if namespace is not None:
                ns_pth = tgt_dir + "/" + namespace
                if not exists(ns_pth):
                    mkdir(ns_pth)

                init_namespace_dir(ns_pth, cfg)
                tgt_pth = ns_pth + "/" + tgt_name

        if isdir(src_pth):
            hmap.update(_rg_dir(src_pth, tgt_name, tgt_pth, cfg, overwrite_file))
        else:
            bhash = _rg_file(src_pth, tgt_name, tgt_pth, cfg, overwrite_file)
            if bhash is not None:
                hmap[pth_as_key(tgt_pth)] = bhash

    return hmap



# from pkglts.templating import same
#
# try:
#     from .github_versioning import fetch_github_version
# except ImportError:
#     fetch_github_version = same
#
#
# mapping = {'version.fetch_github': fetch_github_version}



from pkglts.config_management import Config
from pkglts.option.pysetup.config import check, require, update_parameters


def test_update_parameters():
    cfg = {}
    update_parameters(cfg)
    assert len(cfg['pysetup']) == 2


def test_config_check_intended_version_exists():
    cfg = Config(dict(pysetup={'intended_versions': [], 'require': []}))
    assert 'pysetup.intended_versions' in check(cfg)
    assert 'pysetup.require' not in check(cfg)

    cfg = Config(dict(pysetup={'intended_versions': ["27"],
                               'require': [{'pkg_mng': 'walou', 'name': 'numpy'}]}))
    assert 'pysetup.require' in check(cfg)


def test_require():
    cfg = Config(dict(test={},
                      pysetup={'intended_versions': ["27"],
                               'require': []}))

    assert len(require('option', cfg)) == 5
    assert len(require('setup', cfg)) == 0
    assert len(require('install', cfg)) == 0
    assert len(require('dvlpt', cfg)) == 0



# from pkglts.templating import (get_comment_marker, parse,
#                                reconstruct_txt_div, replace,
#                                swap_divs)
#
#
# def upper(txt, env):
#     del env
#     return txt.upper()
#
#
# def use_env(txt, env):
#     del txt
#     return env['toto']
#
#
# def test_replace_handle_no_div():
#     txt = "print 'toto'"
#     new_txt = replace(txt, {}, None)
#     assert txt == new_txt
#
#
# def test_replace_handle_bracket_end():
#     txt = "d = {'toto': 1}"
#     new_txt = replace(txt, {}, None)
#     assert txt == new_txt
#
#     txt = "d = {'toto': 1{"
#     new_txt = replace(txt, {}, None)
#     assert txt == new_txt
#
#
# def test_replace_handle_unknown_div_class():
#     txt = """
# print 'before'
# # {{,
# print 'titi'
# # }}
# print 'after'
# """
#     new_txt = replace(txt, {}, None)
#     assert new_txt == "\nprint 'before'\nprint 'titi'\nprint 'after'\n"
#
#
# def test_replace_handle_non_commented_lines():
#     txt = """
# print 'before'
# {{,
# print 'titi'
# }}
# print 'after'
# """
#     new_txt = replace(txt, {}, None)
#     assert new_txt == "\nprint 'before'\n\nprint 'titi'\n\nprint 'after'\n"
#
#
# def test_replace_handle_inline_div():
#     txt = "print #{{, 'titi'}}"
#     new_txt = replace(txt, {}, None)
#     assert new_txt == "print 'titi'"
#
#
# def test_replace_handle_nested_div():
#     txt = "print '{{upper, titi {{, toto}} retiti}}'"
#     handlers = {'upper': upper}
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == "print 'TITI TOTO RETITI'"
#
#
# def test_replace_handle_multi_class():
#     txt = "print '{{toto upper, toto}}'"
#     handlers = {'upper': upper}
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == "print 'TOTO'"
#
#
# def test_replace_handle_remove_class():
#     txt = "print 'toto{{toto remove, titi}}'"
#     handlers = {'upper': upper}
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == "print 'toto'"
#
#
# def test_replace_handle_delete_class():
#     txt = "print 'toto{{toto del, titi}}'"
#     handlers = {}
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == "print 'toto_'"
#
#
# def test_replace_handle_existing_inferior_sign_in_file():
#     txt = "print '<toto>'{{remove, titi}}"
#     handlers = {'upper': upper}
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == "print '<toto>'"
#
#
# def test_replace_pass_env_to_handlers():
#     txt = "print '{{use_env, }}'"
#     handlers = {'use_env': use_env}
#     new_txt = replace(txt, handlers, {'toto': 'new toto'})
#     assert new_txt == "print 'new toto'"
#
#
# def test_replace_preserve_indentation():
#     txt = "print 'toto'\n\t# {{upper, toto}}"
#     handlers = {'upper': upper}
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == "print 'toto'\n\tTOTO"
#
#
# def test_replace_preserve_upstream_fmt1():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
# # {{upper, inside = 1}}
# after = 1
# """
#     res = """
# before = 1
# INSIDE = 1
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_replace_preserve_upstream_fmt2():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
#
# # {{upper, inside = 1}}
# after = 1
# """
#     res = """
# before = 1
#
# INSIDE = 1
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_replace_preserve_upstream_fmt3():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
# # {{upper,
# inside = 1
# # }}
# after = 1
# """
#     res = """
# before = 1
# INSIDE = 1
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_replace_preserve_upstream_fmt4():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
#
# # {{upper,
# inside = 1
# # }}
# after = 1
# """
#     res = """
# before = 1
#
# INSIDE = 1
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_replace_preserve_upstream_fmt5():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
#
# # {{upper, inside = 1}}
# after = 1
# """
#     res = """
# before = 1
#
# INSIDE = 1
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_replace_preserve_upstream_fmt6():
#     handlers = {'upper': upper}
#
#     txt = """
#  {{upper, inside = 1}}
# after = 1
# """
#     res = """
#  INSIDE = 1
# after = 1
# """
#     new_txt = replace(txt, handlers, None, ".. ")
#     assert new_txt == res
#
#
# def test_replace_preserve_downstream_fmt1():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
# # {{upper, inside = 1}}
# after = 1
# """
#     res = """
# before = 1
# INSIDE = 1
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_replace_preserve_downstream_fmt2():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
# # {{upper, inside = 1}}
#
# after = 1
# """
#     res = """
# before = 1
# INSIDE = 1
#
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_replace_preserve_downstream_fmt3():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
# # {{upper,
# inside = 1
# # }}
# after = 1
# """
#     res = """
# before = 1
# INSIDE = 1
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_replace_preserve_downstream_fmt4():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
# # {{upper,
# inside = 1
# # }}
#
# after = 1
# """
#     res = """
# before = 1
# INSIDE = 1
#
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_replace_preserve_downstream_fmt5():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
# # {{upper,
# inside = 1
#
# # }}
# after = 1
# """
#     res = """
# before = 1
# INSIDE = 1
#
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_replace_preserve_downstream_fmt6():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
# # {{upper,
#
# inside = 1
#
# # }}
# after = 1
# """
#     res = """
# before = 1
#
# INSIDE = 1
#
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_replace_preserve_spacing():
#     txt = "title\n=====\n\n{{upper, toto}}\n\n"
#     handlers = {'upper': upper}
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == "title\n=====\n\nTOTO\n\n"
#
#
# def test_replace_builtin_func_key():
#     txt = "print '{{key, toto}}'"
#     handlers = {}
#     env = {'toto': 'aaaa'}
#     new_txt = replace(txt, handlers, env)
#     assert new_txt == "print 'aaaa'"
#
#
# def test_replace_builtin_func_key_nested():
#     txt = "print '{{key, toto.titi}}'"
#     handlers = {}
#     env = {'toto.titi': 'aaaa', 'toto': {'titi': 'bbbb'}}
#     new_txt = replace(txt, handlers, env)
#     assert new_txt == "print 'bbbb'"
#
#
# def test_replace_builtin_func_unknown_key():
#     txt = "print '{{key, toto}}'"
#     new_txt = replace(txt, {}, {})
#     assert new_txt == "print 'toto'"
#
#     txt = "print '{{key, toto.titi}}'"
#     new_txt = replace(txt, {}, {})
#     assert new_txt == "print 'toto.titi'"
#
#
# def test_replace_keep_template_fmt_in_pkglts_div():
#     txt = """
# print("toto")
# # {{pkglts,
# print("titi")
# # }}
# print("tutu")
# """
#     new_txt = replace(txt, {}, {})
#     assert new_txt == txt
#
#
# def test_replace_keep_template_fmt_in_pkglts_div_inline():
#     txt = """
# print{{pkglts, ("titi")}}
# """
#     new_txt = replace(txt, {}, {})
#     assert new_txt == txt
#
#
# def test_template_pkglts_divs_can_be_reformatted():
#     txt = """
# print("toto")
# # {{pkglts,
# print("titi")
# # }}
# print("tutu")
# """
#     new_txt = replace(txt, {}, {})
#     new_txt2 = replace(new_txt, {}, {})
#     assert new_txt2 == new_txt
#
#
# def test_template_pkglts_divs_can_be_reformatted2():
#     txt = """
# print("toto")
# # {{pkglts modif,
# print("titi")
# # }}
# print("tutu")
# """
#
#     def modif(txt, env):
#         return txt.upper()
#
#     new_txt = replace(txt, dict(modif=modif), {})
#
#     def modif(txt, env):
#         return txt.lower()
#
#     new_txt2 = replace(new_txt, dict(modif=modif), {})
#     assert new_txt2 != new_txt
#
#
# def test_template_handle_different_types_of_comments_inline():
#     txt = "print#{{div, ('titi')}}"
#     new_txt = replace(txt, {}, {}, comment_marker="#")
#     assert new_txt == "print('titi')"
#
#     txt = "print%{{div, ('titi')}}"
#     new_txt = replace(txt, {}, {}, comment_marker="%")
#     assert new_txt == "print('titi')"
#
#     txt = "before\n.. {{div, inline}}\n\nafter"
#     new_txt = replace(txt, {}, {}, comment_marker=".. ")
#     assert new_txt == "before\ninline\n\nafter"
#
#
# def test_template_handle_different_types_of_comments_div():
#     txt = "before = 1\n# {{div,\ninside = 1\n# }}\nafter = 1"
#     new_txt = replace(txt, {}, {}, comment_marker="#")
#     assert new_txt == "before = 1\ninside = 1\nafter = 1"
#
#     txt = "before = 1\n% {{div,\ninside = 1\n% }}\nafter = 1"
#     new_txt = replace(txt, {}, {}, comment_marker="%")
#     assert new_txt == "before = 1\ninside = 1\nafter = 1"
#
#     txt = "before = 1\n.. {{div,\n\ninside = 1\n.. }}\n\nafter = 1"
#     new_txt = replace(txt, {}, {}, comment_marker=".. ")
#     assert new_txt == "before = 1\n\ninside = 1\n\nafter = 1"
#
#
# def test_template_pkglts_upstream_fmt1():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
# # {{pkglts upper, inside = 1}}
# after = 1
# """
#     res = """
# before = 1
# # {{pkglts upper, INSIDE = 1}}
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_template_pkglts_upstream_fmt2():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
#
# # {{pkglts upper, inside = 1}}
# after = 1
# """
#     res = """
# before = 1
#
# # {{pkglts upper, INSIDE = 1}}
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_template_pkglts_upstream_fmt3():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
# # {{pkglts upper,
# inside = 1
# # }}
# after = 1
# """
#     res = """
# before = 1
# # {{pkglts upper,
# INSIDE = 1
# # }}
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_template_pkglts_upstream_fmt4():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
#
# # {{pkglts upper,
# inside = 1
# # }}
# after = 1
# """
#     res = """
# before = 1
#
# # {{pkglts upper,
# INSIDE = 1
# # }}
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_template_pkglts_downstream_fmt1():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
# # {{pkglts upper, inside = 1}}
#
# after = 1
# """
#     res = """
# before = 1
# # {{pkglts upper, INSIDE = 1}}
#
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_template_pkglts_downstream_fmt2():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
# # {{pkglts upper, inside = 1 }}
# after = 1
# """
#     res = """
# before = 1
# # {{pkglts upper, INSIDE = 1 }}
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_template_pkglts_downstream_fmt3():
#     handlers = {'upper': upper}
#
#     txt = """
# before = 1
# # {{pkglts upper,
# inside = 1
#
# # }}
# after = 1
# """
#     res = """
# before = 1
# # {{pkglts upper,
# INSIDE = 1
#
# # }}
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_template_pkglts_inline_to_div1():
#     def upper_div(txt, env):
#         return "\n" + txt.upper() + "\n"
#
#     handlers = {'upper': upper_div}
#
#     txt = """
# before = 1
# # {{pkglts upper, inside = 1}}
# after = 1
# """
#     res = """
# before = 1
# # {{pkglts upper,
# INSIDE = 1
# # }}
# after = 1
# """
#     new_txt = replace(txt, handlers, None)
#     assert new_txt == res
#
#
# def test_get_comment_marker():
#     assert get_comment_marker("toto.py") == "#"
#     assert get_comment_marker("toto.ini") == "#"
#     assert get_comment_marker("toto.cfg") == "#"
#     assert get_comment_marker("toto.yml") == "#"
#     assert get_comment_marker("toto.rst") == ".. "
#     assert get_comment_marker("toto.bat") == "REM "
#
#
# def test_reconstruct_txt():
#     txt = """
# before = 1  # {{upper, inline = 1}}
# # {{lower,
# div = 1
# # }}
# after = 1
# """
#     root = parse(txt, "#")
#     new_txt = reconstruct_txt_div(root)
#     assert new_txt == txt
#
#
# def test_swap_divs():
#     src_txt = """
# before = 1  # {{pkglts upper, inline = 1}}{{pkglts up2, inline = 3}}
# # {{pkglts lower,
# div = 1
# # }}
# after = 1
# # {{pkglts, inline = 2}}
# after2 = 1
# """
#     tgt_txt = """
# before = 1  # {{pkglts upper, None}}{{pkglts up2, None}}
# # {{pkglts lower,
# None
# # }}
# after = 1
# # {{pkglts, None}}
# after2 = 1
# """
#     new_txt = swap_divs(src_txt, tgt_txt, "#")
#     assert new_txt == src_txt
#
#
# def test_swap_divs_returns_none_if_some_pkglts_divs_missing():
#     src_txt = """
# before = 1  # {{pkglts upper, inline = 1}}{{pkglts up2, inline = 3}}
# # {{pkglts lower,
# div = 1
# # }}
# after = 1
# # {{pkglts, inline = 2}}
# after2 = 1
# """
#     tgt_txt = """
# before = 1  # {{pkglts upper, None}}
# # {{pkglts lower,
# None
# # }}
# after = 1
# # {{pkglts, None}}
# after2 = 1
# """
#     txt = swap_divs(src_txt, tgt_txt, "#")
#     assert txt is None
#
#
# def test_template_core_handler_get_key_return_str():
#     src_txt = "{{key, base.param}}"
#     cfg = dict(base={'param': 0})
#     tgt_txt = replace(src_txt, {}, cfg)
#
#     assert tgt_txt == "0"



"""
Set of function used to regenerate html files from notebooks.
"""
import fnmatch
import os
import os.path
import shutil

import nbconvert


def write_rst_file_with_resources(body, resources):
    """Helper function.
    """
    # Create folder if the path not exists
    if not os.path.exists(resources["metadata"]["path"]):
        os.makedirs(resources["metadata"]["path"])

    # Keep resources metadata like image in the rst
    writer = nbconvert.writers.FilesWriter()
    writer.write(body, resources, notebook_name=resources["metadata"]["name"])


def write_rst_file_with_filename(body, filename):
    """
    Warning, This write not save image in the transformation to rst format.
    """
    # Create folder if the path not exists
    if not os.path.exists(os.path.dirname(filename)):
        os.makedirs(os.path.dirname(filename))

    # Write rst file
    with open(filename, "w+") as rst_file:
        rst_file.write(body)


def find_notebook_file(root_directory):
    """Get recursively all notebook filenames in the src_directory.

    Args:
        root_directory (str): base directory to search

    Returns:
        (list of str)
    """
    matches = []
    for root, _, filenames in os.walk(root_directory):
        for filename in fnmatch.filter(filenames, '*.ipynb'):

            # Avoid tmp notebook in hidden folder like *-checkpoint.ipynb
            if os.path.basename(root):
                if os.path.basename(root)[0] == '.':
                    continue

            matches.append(os.path.join(root, filename))

    return matches


def main(cfg, target=".", overwrite=False):
    """Main function called to walk the package

    Args:
        cfg (Config):  current package configuration
        target (str): place to write plugin def into
        overwrite (bool): whether or not to overwrite previous definition
                          files. Default to False.
    """
    del target
    del overwrite

    src_directory = cfg["notebook"]['src_directory']
    len_src_directory = len(src_directory)

    nb_filenames = find_notebook_file(src_directory)

    # define documentation rst notebook directory
    dst_rst_directory = os.path.join("doc", "_notebook")

    # remove previous folder
    if os.path.exists(dst_rst_directory):
        shutil.rmtree(dst_rst_directory)

    index_body = """\
========
Notebook
========

.. toctree::
    :glob:
    :caption: Notebook

"""

    rst_exporter = nbconvert.RSTExporter()
    for nb_filename in nb_filenames:
        # Convert each notebook to rst
        body, resources = rst_exporter.from_filename(nb_filename)

        # Remove basename src_directory in the path
        resources["metadata"]["path"] = \
            resources["metadata"]["path"][len_src_directory + 1:]

        # Get the local file_path of the notebook write
        local_file_path = os.path.join(resources["metadata"]["path"],
                                       resources["metadata"]["name"])

        # Add dst_rst_directory in the path
        resources["metadata"]["path"] = os.path.join(
            dst_rst_directory, resources["metadata"]["path"])

        # Write rst with this resources
        write_rst_file_with_resources(body, resources)

        # Save the notebook rst position in the index body
        index_body += "    " + local_file_path.replace("\\", "/") + "\n"

    # Write rst_index body in the root src directory
    rst_index_filename = os.path.join(dst_rst_directory, "index.rst")
    write_rst_file_with_filename(index_body, rst_index_filename)



"""
Set of function related to handling the configuration of this option.
"""
from pkglts.dependency import Dependency


def require(purpose, cfg):
    """List of requirements for this option for a given purpose.

    Args:
        purpose (str): either 'option', 'setup', 'install' or 'dvlpt'
        cfg (Config):  current package configuration

    Returns:
        (list of Dependency)
    """
    if purpose == 'option':
        options = ['test']
        return [Dependency(name) for name in options]

    if purpose == 'dvlpt':
        deps = [Dependency('coverage')]
        test_suite = cfg['test']['suite_name']
        if test_suite == 'pytest':
            deps.append(Dependency('pytest-cov'))

        return deps

    return []



from os import listdir, mkdir
from os.path import exists
import pytest

from pkglts.manage import clean

from .small_tools import ensure_created, rmdir


@pytest.fixture()
def tmp_dir():
    pth = 'toto_manage_clean'
    ensure_created(pth)

    yield pth

    rmdir(pth)


def test_clean_remove_pyc_files(tmp_dir):
    name = tmp_dir + "/" + "toto.pyc"
    with open(name, 'w') as f:
        f.write("toto")

    clean(tmp_dir)
    assert not exists(name)


def test_clean_remove_pycache_directories(tmp_dir):
    pycache = tmp_dir + "/" + "__pycache__"
    mkdir(pycache)
    name = pycache + "/" + "toto.py"
    with open(name, 'w') as f:
        f.write("toto")

    clean(tmp_dir)
    assert not exists(pycache)
    assert not exists(name)


def test_clean_do_not_remove_py_files(tmp_dir):
    name = tmp_dir + "/" + "toto.py"
    with open(name, 'w') as f:
        f.write("toto")

    clean(tmp_dir)
    assert exists(name)


def test_clean_do_not_remove_hidden_files(tmp_dir):
    name = tmp_dir + "/" + ".toto.pyc"
    with open(name, 'w') as f:
        f.write("toto")

    clean(tmp_dir)
    assert exists(name)


def test_clean_do_not_explore_hidden_directories(tmp_dir):
    hidden = tmp_dir + "/" + ".test"
    mkdir(hidden)
    name = hidden + "/" + "toto.py"
    with open(name, 'w') as f:
        f.write("toto")

    clean(tmp_dir)
    assert exists(hidden)
    assert exists(name)


def test_clean_do_not_explore_clean_no_directories(tmp_dir):
    hidden = tmp_dir + "/" + "test"
    mkdir(hidden)
    name = hidden + "/" + "toto.py"
    with open(name, 'w') as f:
        f.write("toto")

    with open(hidden + "/" + "clean.no", 'w') as f:
        f.write("")

    clean(tmp_dir)
    assert exists(hidden)
    assert exists(name)


def test_clean_remove_dist_build_directories(tmp_dir):
    for name in ("dist", "build"):
        mkdir(tmp_dir + "/" + name)

    clean(tmp_dir)
    assert len(listdir(tmp_dir)) == 0



"""
Set of function to extend jinja2 environment.
"""
from pkglts.local import pkg_full_name, src_dir


def environment_extensions(cfg):
    """Add more functionality to an environment.

    Args:
        cfg (Config):  current package configuration

    Returns:
        dict of str: any
    """
    return {"pkg_full_name": pkg_full_name(cfg),
            "src_pth": src_dir(cfg)}



import pytest

from pkglts.config_management import Config


def test_generate():
    cfg = Config(dict(license=dict(
        name='mit',
        year="2015",
        organization="org",
        project="project"))
    )
    cfg.load_extra()
    assert len(cfg._env.globals['license'].full_text) > 0


def test_generate_raise_error_if_license_do_not_exists():
    cfg = dict(name='tugudu',
               year="2015",
               organization="org",
               project="project")
    with pytest.raises(IOError):
        Config(dict(license=cfg)).load_extra()



from pkglts.config_management import Config


def test_src_pth():
    cfg = Config({'base': {'pkgname': 'toto', 'namespace': None}})
    cfg.load_extra()
    assert cfg._env.globals['base'].src_pth == "src/toto"
#     cfg = Config({'base': {'pkgname': 'toto', 'namespace': 'oa'}})
#     cfg.load_extra()
#     assert cfg._env.globals['base'].src_pth == "src/oa/toto"
#
#
# def test_pkg_full_name():
#     cfg = Config({'base': {'pkgname': 'toto', 'namespace': None}})
#     cfg.load_extra()
#     assert cfg._env.globals['base'].pkg_full_name == "toto"
#     cfg = Config({'base': {'pkgname': 'toto', 'namespace': 'oa'}})
#     cfg.load_extra()
#     assert cfg._env.globals['base'].pkg_full_name == "oa.toto"



"""
Main config object and functions to manipulate it.
"""
import json
import logging
from datetime import date
from os.path import join as pj

from jinja2 import Environment, StrictUndefined, UndefinedError

from .config import pkg_cfg_file, pkglts_dir
from .option.git.config import update_parameters as git_update_parameters
from .option.pypi.config import update_parameters as pypi_update_parameters
from .option_tools import available_options, find_available_options

try:
    string_type = basestring
except NameError:
    string_type = str

CURRENT_PKG_CFG_VERSION = 11

LOGGER = logging.getLogger(__name__)

DEFAULT_CFG = dict(_pkglts=dict(use_prompts=False,
                                auto_install=True,
                                install_front_end='stdout',
                                version=CURRENT_PKG_CFG_VERSION))

find_available_options()


class ConfigSection(object):
    """Small class to allow accessing parameters using the dot
    method instead of ['param_name'] method
    """
    pass


class Config(dict):
    """Object used to store both a templated version of the config as a dict interface
    its resolution and a jinja2 environment that reflect the config.
    """

    def __init__(self, *args, **kwds):
        dict.__init__(self)
        self._tpl = dict(*args, **kwds)

        # initialise associated Jinja2 environment
        self._env = Environment(undefined=StrictUndefined)
        self._env.keep_trailing_newline = True

        # add global filters and test
        self._env.globals['today'] = lambda: date.today().isoformat()

        self.add_test('available', self._is_available)

        # resolve
        self.resolve()

    def template(self):
        """Associated template even after resolution."""
        return self._tpl

    def add_test(self, name, func):
        """Add a new test in jinja2 environment.

        Args:
            name (str): name of test (must be unique)
            func (callable): function use for test

        Returns:
            None
        """
        self._env.tests[name] = func

    def _is_available(self, opt_name):
        return opt_name in self

    def _add_param(self, opt_name, param_name, param_value):
        """Add a new parameter value in the config

        Args:
            opt_name (str): Option name
            param_name (str): parameter name
            param_value (any): parameter value

        Returns:
            None
        """
        self[opt_name][param_name] = param_value
        setattr(self._env.globals[opt_name], param_name, param_value)

    def resolve(self):
        """Try to resolve all templated items.

        Returns:
            None
        """
        to_eval = []
        for opt_name, cfg in self._tpl.items():
            self[opt_name] = {}
            self._env.globals[opt_name] = ConfigSection()
            for key, param in cfg.items():
                if isinstance(param, string_type):
                    to_eval.append((opt_name, key, param))
                else:
                    self._add_param(opt_name, key, param)

        nb_iter_max = len(to_eval) ** 2
        cur_iter = 0
        while to_eval and cur_iter < nb_iter_max:
            cur_iter += 1
            opt_name, key, param = to_eval.pop(0)
            try:
                txt = self.render(param)
                self._add_param(opt_name, key, txt)
            except UndefinedError:
                to_eval.append((opt_name, key, param))

        if to_eval:
            msg = "unable to fully render config\n"
            for item in to_eval:
                msg += "%s:%s '%s'\n" % item
            raise UserWarning(msg)

    def load_extra(self):
        """load option specific handlers.

        Returns:
            None
        """
        for opt_name in self:
            if not opt_name.startswith("_"):
                try:
                    opt = available_options[opt_name]
                    for func_name, func in opt.environment_extensions(self).items():
                        setattr(self._env.globals[opt_name], func_name, func)
                except KeyError:
                    raise KeyError("option '%s' does not exists" % opt_name)

    def installed_options(self):
        """List all installed options.

        Returns:
            (iter of str)
        """
        for key in self:
            if not key.startswith("_"):
                yield key

    def render(self, txt):
        """Use items in config to render text

        Args:
            txt (str): templated text to render

        Returns:
            (str): same text where all templated parts have been replaced
                   by their values.
        """
        return self._env.from_string(txt).render()


def get_pkg_config(rep="."):
    """Read pkg_cfg file associated to this package.

    Args:
        rep (str): directory to search for info

    Returns:
        (Config): Config initialized with pkg_config
    """
    with open(pj(rep, pkglts_dir, pkg_cfg_file), 'r') as fhr:
        pkg_cfg = json.load(fhr)

    # update version of pkg_config
    file_version = pkg_cfg['_pkglts'].get('version', 0)
    for i in range(file_version, CURRENT_PKG_CFG_VERSION):
        upgrade_pkg_cfg_version(pkg_cfg, i)

    # create Config object
    cfg = Config(pkg_cfg)
    cfg.load_extra()

    # write back config if version has been updated
    if file_version < CURRENT_PKG_CFG_VERSION:
        write_pkg_config(cfg, rep)

    return cfg


def write_pkg_config(cfg, rep="."):
    """Store config associated to this package on disk.

    Args:
        cfg (Config): current working config
        rep (str): directory to search for info

    Returns:
        None
    """
    LOGGER.info("write package config")
    pkg_cfg = dict(cfg.template())

    with open(pj(rep, pkglts_dir, pkg_cfg_file), 'w') as fhw:
        json.dump(pkg_cfg, fhw, sort_keys=True, indent=2)


def upgrade_pkg_cfg_version(pkg_cfg, version):
    """Upgrade the version of pkg_cfg file from version to version +1

    Args:
        pkg_cfg (dict of str, any): package configuration
        version (int): current version of file

    Returns:
        (dict of str: any): a reference to an updated pkg_cfg
    """
    if version == 0:
        pkg_cfg['_pkglts']['version'] = 1
    elif version == 1:
        pkg_cfg['_pkglts']['version'] = 2
    elif version == 2:
        pkg_cfg['_pkglts']['version'] = 3
        if 'pysetup' in pkg_cfg:
            section = pkg_cfg['pysetup']
            section['require'] = section.get('require', [])
    elif version == 3:
        pkg_cfg['_pkglts']['version'] = 4
        if 'test' in pkg_cfg:
            section = pkg_cfg['test']
            section['suite_name'] = section.get('suite_name', "pytest")
    elif version == 4:
        pkg_cfg['_pkglts']['version'] = 5
        if 'base' in pkg_cfg:
            section = pkg_cfg['base']
            section['namespace_method'] = section.get('namespace_method', "pkg_util")
    elif version == 5:
        pkg_cfg['_pkglts']['version'] = 6
        if 'pysetup' in pkg_cfg:
            section = pkg_cfg['pysetup']
            deps = []
            for pkg_mng, name in section.get('require', []):
                if pkg_mng == 'none':
                    deps.append(dict(name=name))
                else:
                    deps.append(dict(name=name, pkg_mng=pkg_mng))
            section['require'] = deps
    elif version == 6:
        pkg_cfg['_pkglts']['version'] = 7
        if 'sphinx' in pkg_cfg:
            section = pkg_cfg['sphinx']
            section['build_dir'] = section.get('build_dir', "build/sphinx")
    elif version == 7:
        pkg_cfg['_pkglts']['version'] = 8
        if 'doc' in pkg_cfg:
            section = pkg_cfg['doc']
            section['fmt'] = section.get('fmt', 'rst')
    elif version == 8:
        pkg_cfg['_pkglts']['version'] = 9
        if 'github' in pkg_cfg or 'gitlab' in pkg_cfg:
            LOGGER.info("\n"
                        "################################################\n"
                        "\n"
                        "\n"
                        "please remove '.gitignore file and regenerate package\n"
                        "\n"
                        "\n"
                        "################################################\n")
            git_update_parameters(pkg_cfg)
    elif version == 9:
        pkg_cfg['_pkglts']['version'] = 10
        if 'pypi' in pkg_cfg:
            LOGGER.info("\n"
                        "################################################\n"
                        "\n"
                        "\n"
                        "please remove '.pypirc file and regenerate package\n"
                        "\n"
                        "\n"
                        "################################################\n")
            # save section
            mem = dict(pkg_cfg["pypi"])
            # get newly defined list of servers
            pypi_update_parameters(pkg_cfg)
            servers = dict(pkg_cfg["pypi"]["servers"])
            # update option
            pkg_cfg["pypi"] = mem
            pkg_cfg["pypi"]["servers"] = pkg_cfg["pypi"].get("servers", {})
            pkg_cfg["pypi"]["servers"].update(servers)
    elif version == 10:
        pkg_cfg['_pkglts']['version'] = 11
        if 'data' in pkg_cfg:
            section = pkg_cfg['data']
            section['filetype'] = section.get('filetype', [".json", ".ini"])
            section['use_ext_dir'] = section.get('use_ext_dir', False)

    return pkg_cfg



"""
Set of function to extend jinja2 environment.
"""
from pkglts.dependency import Dependency
from pkglts.option_tools import available_options


def requirements(cfg, requirement_name):
    """Check all requirements for installed options.

    Args:
        cfg (Config):  current package configuration
        requirement_name (str): type of requirement 'install', 'dvlpt'

    Returns:
        (list of str): list of required packages names
    """
    reqs = {}
    for name in cfg.installed_options():
        try:
            opt = available_options[name]
            for dep in opt.require(requirement_name, cfg):
                reqs[dep.name] = dep
        except KeyError:
            raise KeyError("option '%s' does not exists" % name)

    return [reqs[name] for name in sorted(reqs)]


def pkg_url(cfg):
    """Extract a valid url from all config sections.

    Args:
        cfg (Config):  current package configuration

    Returns:
        (str): a valid url for the package
    """
    try:
        url = cfg['base']['url']
        if url is not None:
            return url
    except KeyError:
        pass

    try:
        url = cfg['github']['url']
        if url is not None:
            return url
    except KeyError:
        pass

    try:
        url = cfg['gitlab']['url']
        if url is not None:
            return url
    except KeyError:
        pass

    try:
        url = cfg['pypi']['url']
        if url is not None:
            return url
    except KeyError:
        pass

    try:
        url = cfg['readthedocs']['url']
        if url is not None:
            return url
    except KeyError:
        pass

    return ""


def environment_extensions(cfg):
    """Add more functionality to an environment.

    Args:
        cfg (Config):  current package configuration

    Returns:
        dict of str: any
    """
    req_install = requirements(cfg, 'install')
    for dep in cfg['pysetup']['require']:
        req_install.append(Dependency(**dep))

    req_dvlpt = requirements(cfg, 'dvlpt')

    def req(name):
        """For internal use only."""
        if name == 'install':
            return req_install
        elif name == 'dvlpt':
            return req_dvlpt
        else:
            raise UserWarning("WTF")

    cfg.add_test('is_pip_dep', Dependency.is_pip)

    return {"pkg_url": pkg_url(cfg),
            "requirements": req}



"""
Set of function to extend jinja2 environment.
"""
from pkglts.option.doc import fmt_badge


def environment_extensions(cfg):
    """Add more functionality to an environment.

    Args:
        cfg (Config):  current package configuration

    Returns:
        dict of str: any
    """
    owner = cfg['github']['owner']
    project = cfg['github']['project']

    url = "travis-ci.org/%s/%s" % (owner, project)
    img = url + ".svg?branch=master"
    badge = fmt_badge(img, url, "Travis build status", cfg['doc']['fmt'])

    return {"badge": badge}



# from os.path import exists
# import sys
#
#
# def list_python_envs(versions):
#     """ Check the system for all available python distributions
#
#     args:
#      - versions (list of str): list of python versions
#                            (e.g. '27', '34') to check
#     """
#     if 'win' in sys.platform:
#         installed = []
#         # check known locations
#         for pyver in versions:
#             if exists("C:/Python%s/python.exe" % pyver):
#                 # TODO check alternative location
#                 installed.append("py%s" % pyver)
#
#         return installed
#     else:  # use which command
#         print("need some linux to try")
#         return []



from os import mkdir, rmdir
from os.path import exists
import pytest

from pkglts.config_management import Config
from pkglts.option.notebook.config import check, require, update_parameters


@pytest.fixture()
def tmp_dir():
    pth = "nb_test_config"
    mkdir(pth)

    yield pth

    if exists(pth):
        rmdir(pth)


def test_update_parameters():
    cfg = {}
    update_parameters(cfg)
    assert len(cfg['notebook']) == 1


def test_config_check_src_directory(tmp_dir):
    cfg = Config(dict(notebook={'src_directory': "failed_nb"}))
    assert 'notebook.src_directory' in check(cfg)

    cfg = Config(dict(notebook={'src_directory': tmp_dir}))
    assert 'notebook.src_directory' not in check(cfg)


def test_require():
    cfg = Config(dict(notebook={}))

    assert len(require('option', cfg)) == 1
    assert len(require('setup', cfg)) == 0
    assert len(require('install', cfg)) == 0
    assert len(require('dvlpt', cfg)) == 1



import json
from os import remove
from os.path import exists
from os.path import join as pj
import pytest

from pkglts.config import pkglts_dir, pkg_cfg_file
from pkglts.config_management import Config, get_pkg_config, write_pkg_config
from pkglts.manage import (init_pkg, regenerate_package)

from .small_tools import ensure_created, rmdir


def addendum(init_file):
    """ modify init_file in first pkglts div
    """
    with open(init_file, 'r') as f:
        lines = f.read().splitlines()
    
    lines.insert(1, "addendum")
    
    with open(init_file, 'w') as f:
        f.write("\n".join(lines))


@pytest.fixture()
def tmp_pths():
    pth = 'toto_mg_rg'
    ensure_created(pth)
    init_pkg(pth)
    with open(pj(pth, pkglts_dir, pkg_cfg_file), 'r') as f:
        pkg_cfg = json.load(f)
    
    pkg_cfg['base'] = dict(pkgname='toto', namespace=None,
                           authors=[('moi', 'moi@email.com')],
                           namespace_method="pkg_util",
                           url=None)
    cfg = Config(pkg_cfg)
    write_pkg_config(cfg, pth)
    regenerate_package(cfg, pth)
    
    init_file = pj(pth, "src", "toto", "__init__.py")
    
    yield pth, init_file
    
    rmdir(pth)


def test_regenerate_pass(tmp_pths):
    tmp_dir, init_file = tmp_pths
    with open(pj(tmp_dir, pkglts_dir, pkg_cfg_file), 'r') as f:
        pkg_cfg = json.load(f)
    
    pkg_cfg['base'] = dict(pkgname='toto', namespace=None,
                           authors=[('moi', 'moi@email.com')],
                           namespace_method="pkg_util",
                           url=None)
    cfg = Config(pkg_cfg)
    regenerate_package(cfg, tmp_dir)
    assert exists(init_file)


def test_regenerate_check_pkg_cfg_validity(tmp_pths):
    tmp_dir, init_file = tmp_pths
    with open(pj(tmp_dir, pkglts_dir, pkg_cfg_file), 'r') as f:
        pkg_cfg = json.load(f)
    
    pkg_cfg['base'] = dict(pkgname='1toto', namespace=None,
                           authors=[('moi', 'moi@email.com')],
                           namespace_method="pkg_util",
                           url=None)
    cfg = Config(pkg_cfg)
    assert not regenerate_package(cfg, tmp_dir)


def test_regenerate_handle_conflicts_keep(tmp_pths, mocker):
    tmp_dir, init_file = tmp_pths
    cfg = get_pkg_config(tmp_dir)
    
    with open(init_file, 'w') as f:
        f.write("modified")
    
    with mocker.patch('pkglts.manage.get_user_permission',
                      return_value=False):
        regenerate_package(cfg, tmp_dir)
    
    with open(init_file, 'r') as f:
        assert f.read() == "modified"


def test_regenerate_handle_conflicts_overwrite(tmp_pths, mocker):
    tmp_dir, init_file = tmp_pths
    cfg = get_pkg_config(tmp_dir)
    
    addendum(init_file)
    
    with mocker.patch('pkglts.manage.get_user_permission',
                      return_value=True):
        regenerate_package(cfg, tmp_dir)
    
    with open(init_file, 'r') as f:
        txt = f.read()
        assert "modified" not in txt


def test_regenerate_handle_global_overwrite(tmp_pths):
    tmp_dir, init_file = tmp_pths
    cfg = get_pkg_config(tmp_dir)
    
    addendum(init_file)
    
    regenerate_package(cfg, tmp_dir, overwrite=True)
    
    with open(init_file, 'r') as f:
        txt = f.read()
        assert "modified" not in txt


def test_regenerate_new_files_do_not_generate_conflicts(tmp_pths):
    tmp_dir, init_file = tmp_pths
    cfg = get_pkg_config(tmp_dir)
    
    new_pth = pj(tmp_dir, "src", "toto", "new_file.py")
    with open(new_pth, 'w') as f:
        f.write("txt = 'addendum'")
    
    regenerate_package(cfg, tmp_dir)
    
    assert exists(new_pth)
    with open(new_pth, 'r') as f:
        txt = f.read()
        assert txt == "txt = 'addendum'"


def test_regenerate_remove_user_files_do_not_generate_conflicts(tmp_pths):
    tmp_dir, init_file = tmp_pths
    cfg = get_pkg_config(tmp_dir)
    
    new_pth = pj(tmp_dir, "src", "toto", "new_file.py")
    with open(new_pth, 'w') as f:
        f.write("txt = 'addendum'")
    
    regenerate_package(cfg, tmp_dir)
    
    remove(new_pth)
    
    regenerate_package(cfg, tmp_dir)
    assert not exists(new_pth)


def test_regenerate_fail_if_permanent_section_ids_have_been_modified(tmp_pths):
    tmp_dir, init_file = tmp_pths
    cfg = get_pkg_config(tmp_dir)
    
    with open(init_file, 'a') as f:
        f.write("\n# {# pkglts, test\na = 1\n# #}\n")
    
    with pytest.raises(KeyError):
        regenerate_package(cfg, tmp_dir, overwrite=True)



import pytest

from pkglts.config_management import Config, DEFAULT_CFG
from pkglts.manage_tools import update_opt


print(__file__)


def test_non_existing_option_raises_warning():
    cfg = Config({})
    with pytest.raises(KeyError):
        update_opt('toto', cfg)


def test_option_fetch_parameter_list_from_config():
    cfg = Config(DEFAULT_CFG)
    cfg = update_opt('base', cfg)
    assert 'base' in cfg
    assert 'pkgname' in cfg['base']
    assert 'authors' in cfg['base']


def test_option_handle_no_parameter_list_in_config():
    cfg = Config(DEFAULT_CFG)
    cfg._env.globals['base'] = None
    cfg = update_opt('data', cfg)
    assert 'data' in cfg
    assert len(tuple(cfg['data'].items())) == 2


def test_option_use_default_from_config():
    cfg = Config(DEFAULT_CFG)
    cfg = update_opt('base', cfg)
    assert 'base' in cfg
    assert cfg['base']['authors'][0][0] == 'moi'


# def test_option_prompt_user_if_global_config_ask_for_it(mocker):
#     pkg_cfg = dict(_pkglts={'use_prompts': True})
#
#     with mocker.patch('pkglts.option_tools.loc_input', return_value=''):
#         pkg_cfg = update_opt('base', pkg_cfg)
#         assert 'base' in pkg_cfg
#         cfg = pkg_cfg['base']
#         assert cfg['owner'] == 'moi'


# def test_option_pass_environment_to_config():
#     pkg_cfg = dict(_pkglts={'use_prompts': True})
#     pkg_cfg = update_opt('base', pkg_cfg, extra={'pkg_fullname': 'toto',
#                                                  'owner': 'owner'})
#     assert 'base' in pkg_cfg


# def test_option_register():
#     pkg_cfg = update_opt('base', {}, extra={'pkg_fullname': 'toto',
#                                             'owner': 'owner'})
#     pkg_cfg = update_opt('test', pkg_cfg)
#     assert 'base' in pkg_cfg
#     assert 'test' in pkg_cfg
#
#
# def test_option_look_for_dependencies():
#     extra = {"install_option_dependencies": True,
#              "pkg_fullname": 'toto',
#              'owner': 'moi'}
#     pkg_cfg = update_opt('test', {}, extra=extra)
#     assert 'base' in pkg_cfg
#     assert 'test' in pkg_cfg



from pkglts.config_management import Config
from pkglts.option.base.config import check, require, update_parameters


def test_update_parameters():
    cfg = {}
    update_parameters(cfg)
    assert len(cfg['base']) == 5


def test_config_check_pkg_names():
    for pkg in ('1mypkg', ' mypkg', '1', '1.mypkg',
                ' .mypkg', '.mypkg', 'None.mypkg', 'oa.o.mypkg'):
        cfg = Config(dict(base={'pkgname': pkg,
                                    'namespace': None,
                                    'namespace_method': "pkg_util",
                                    'owner': 'moi',
                                    'url': None}))
        assert 'base.pkgname' in check(cfg)
        cfg = Config(dict(base={'pkgname': 'toto',
                                    'namespace': pkg,
                                    'namespace_method': "pkg_util",
                                    'owner': 'moi',
                                    'url': None}))
        assert 'base.namespace' in check(cfg)

    cfg = Config(dict(base={'pkgname': 'toto',
                                'namespace': None,
                                'namespace_method': "toto",
                                'owner': 'moi',
                                'url': None}))
    assert 'base.namespace_method' in check(cfg)


def test_require():
    cfg = Config(dict(base={}))

    assert len(require('option', cfg)) == 0
    assert len(require('setup', cfg)) == 0
    assert len(require('install', cfg)) == 0
    assert len(require('dvlpt', cfg)) == 0



"""
Base setting for logging in the package.
This module is executed when importing pkglts.
"""
import logging
import logging.handlers
from os import remove, rename
from os.path import exists


def rolling_logs(logger):
    """Associate log files using a rolling mechanism"""
    nb_logs = 5
    tpl = ".pkglts/info.log.%d"
    if exists(tpl % nb_logs):
        remove(tpl % nb_logs)

    for i in range(nb_logs, 0, -1):
        if exists(tpl % (i - 1)):
            rename(tpl % (i - 1), tpl % i)

    if exists(".pkglts/info.log"):
        rename(".pkglts/info.log", tpl % 0)

    info_ch = logging.FileHandler(".pkglts/info.log", 'w')
    logger.addHandler(info_ch)


def main(verbosity):
    """Set logging configuration.

    Args:
        verbosity (int): verbosity level

    Returns:
        None
    """
    if verbosity > 2:
        verbosity = 2
    vlevel = [logging.WARNING, logging.INFO, logging.DEBUG][verbosity]

    fmt = logging.Formatter('%(asctime)s %(levelname)s (%(name)s): %(message)s')

    wng_ch = logging.StreamHandler()
    wng_ch.setLevel(vlevel)
    wng_ch.setFormatter(fmt)

    logger = logging.getLogger("pkglts")
    logger.setLevel(vlevel)
    logger.addHandler(wng_ch)

    try:
        # rolling_logs(logger)
        info_ch = logging.StreamHandler()
        info_ch.setLevel(logging.INFO)
        info_ch.setFormatter(fmt)

        logger.addHandler(info_ch)
    except IOError:
        print("no info handler")


if __name__ == '__main__':
    main(0)



from pkglts.config_management import Config
from pkglts.option.license.config import check, require, update_parameters


def test_update_parameters():
    cfg = {}
    update_parameters(cfg)
    assert len(cfg['license']) == 4


def test_config_check_license_name_exists():
    cfg = Config(dict(license={'name': "", 'year': 2015,
                               'organization': "oa", 'project': "project"}))
    assert 'license.name' in check(cfg)

    cfg = Config(dict(license={'name': "tugudu", 'year': 2015,
                               'organization': "oa", 'project': "project"}))
    assert 'license.name' in check(cfg)


def test_require():
    cfg = Config(dict(license={}))

    assert len(require('option', cfg)) == 1
    assert len(require('setup', cfg)) == 0
    assert len(require('install', cfg)) == 0
    assert len(require('dvlpt', cfg)) == 0



import os
from os import mkdir
from os.path import exists
from shutil import rmtree
from subprocess import call
import sys
from time import sleep


def _activate(name):
    """Internal function used to activate a virtualenv
    """
    if "win" in sys.platform:
        pth = "%s/Scripts/activate_this.py" % name
    else:
        pth = "%s/bin/activate_this.py" % name

    try:  # python27
        execfile(pth, dict(__file__=pth))
    except NameError:  # python3.x
        with open(pth) as f:
            code = compile(f.read(), pth, 'exec')
            exec(code, dict(__file__=pth))


def create_venv(name, mem):
    """Create a new virtualenv with given name and activate it

    .. warning: create a directory with given name in local dir

    ..warning: erase the content of dir 'name' before creating virtualenv

    args:
     - name (str): name of virtualenv to create
     - mem (dict): a place to store values that will be modified
    """
    try:
        if exists(name):
            rmtree(name)
        else:
            mkdir(name)
    except OSError:
        print("unable to create dir")
        sys.exit(0)

    call("virtualenv %s" % name, shell=True)

    # Save state of path
    mem["sys.path"] = tuple(sys.path)
    mem["os.environ['PATH']"] = str(os.environ["PATH"])

    _activate(name)


def clear_venv(name, mem):
    """Remove virtualenv from existence.

    args:
     - name (str): name of virtualenv
     - mem (dict): a place of values that where modified
    """
    # Restore state
    del sys.path[:]
    for item in mem["sys.path"]:
        sys.path.append(item)

    os.environ["PATH"] = str(mem["os.environ['PATH']"])

    # remove directory
    for i in range(5):
        if exists(name):
            try:
                rmtree(name)
                return
            except OSError:
                sleep(0.1)
        else:
            return

    raise OSError("unable to remove directory: %s" % name)



"""
Define actions that can be called with the CLI.
"""
from argparse import ArgumentParser, RawTextHelpFormatter
import json
import logging

from .config_management import (get_pkg_config, write_pkg_config)
from . import logging_tools
from .manage import (clean, init_pkg, install_example_files,
                     regenerate_package, regenerate_option, add_option)
from .tool.history import action_history

LOGGER = logging.getLogger(__name__)


def action_info(*args, **kwds):
    """Display info on package for debug purpose.
    """
    del args  # unused
    del kwds  # unused
    LOGGER.info("package info")
    from pkglts.option_tools import available_options
    print("available_options:")
    for opt_name in available_options:
        print("  ", opt_name)
    cfg = get_pkg_config()
    print("current config (after resolution)")
    for opt_name, opt_params in cfg.items():
        if not opt_name.startswith("_"):
            print(opt_name)
            print(json.dumps(opt_params, sort_keys=True, indent=2))


def action_clean(*args, **kwds):
    """Clean package of all un necessary files.
    """
    del args  # unused
    del kwds  # unused
    LOGGER.info("clean package")
    clean()


def action_init(*args, **kwds):
    """Initialize environment for use of pkglts.
    """
    init_pkg()

    if args:
        action_add(*args, **kwds)


def action_clear(*args, **kwds):
    """Attempt to free the package from pkglts interactions.
    """
    del args  # unused
    del kwds  # unused
    LOGGER.info("clear")
    print("TODO")


def action_update(*args, **kwds):
    """Check if a new version of pkglts is available.
    """
    del args  # unused
    del kwds  # unused
    LOGGER.info("update")
    print("TODO")


def action_regenerate(*args, **kwds):
    """Regenerate all files in the package.
    """
    overwrite = 'overwrite' in kwds

    cfg = get_pkg_config()
    clean()

    if args:
        for name in args:
            LOGGER.info("regenerate '%s'", name)
            regenerate_option(cfg, name, overwrite=overwrite)
    else:
        LOGGER.info("regenerate package")
        regenerate_package(cfg, overwrite=overwrite)


def action_add(*args, **kwds):
    """Add new options in the package.
    """
    del kwds  # unused
    if not args:
        raise UserWarning("need to specify at least one option name")

    LOGGER.info("add option")
    cfg = get_pkg_config()
    for name in args:
        cfg = add_option(name, cfg)

    write_pkg_config(cfg)


def action_remove(*args, **kwds):
    """Remove options from the package.
    """
    del kwds  # unused
    if not args:
        raise UserWarning("need to specify at least one option name")

    LOGGER.info("remove option")
    print("TODO")


def action_example(*args, **kwds):
    """Install example files associated with options.
    """
    del kwds  # unused
    if not args:
        raise UserWarning("need to specify at least one option name")

    LOGGER.info("install examples")
    cfg = get_pkg_config()
    for name in args:
        install_example_files(name, cfg)


def main():
    """Run CLI evaluation"""
    action = dict(
        info=action_info,
        clean=action_clean,
        init=action_init,
        clear=action_clear,
        update=action_update,
        regenerate=action_regenerate,
        rg=action_regenerate,
        add=action_add,
        remove=action_remove,
        example=action_example,
        history=action_history
    )
    # parse argument line
    parser = ArgumentParser(description='Package structure manager',
                            formatter_class=RawTextHelpFormatter)

    act_help = "type of action performed by pmg, one of:\n"
    for name, func in action.items():
        act_help += "\n  - %s: %s" % (name, func.__doc__)

    parser.add_argument('action', metavar='action',
                        choices=tuple(action.keys()),
                        help=act_help)

    parser.add_argument('action_args', nargs='*',
                        help="action to perform on the package")

    parser.add_argument('-e', metavar='extra', nargs=2, action='append',
                        help='extra arguments to pass to the action',
                        dest='extra')

    parser.add_argument("-v", "--verbosity", action="count", default=0,
                        help="increase output verbosity")

    args = parser.parse_args()
    if args.extra is None:
        extra = {}
    else:
        extra = dict(args.extra)

    logging_tools.main(args.verbosity)

    # perform action
    action[args.action](*args.action_args, **extra)


if __name__ == '__main__':
    main()



from pkglts.config_management import Config
from pkglts.option.github.config import check, require, update_parameters


def test_update_parameters():
    cfg = {}
    update_parameters(cfg)
    assert len(cfg['github']) == 3


def test_config_check_project_exists():
    cfg = Config(dict(github={'owner': "", 'project': "", "url": ""}))
    assert 'github.project' in check(cfg)


def test_require():
    cfg = Config(dict(base={}, data={}))

    assert len(require('option', cfg)) == 1
    assert len(require('setup', cfg)) == 0
    assert len(require('install', cfg)) == 0
    assert len(require('dvlpt', cfg)) == 0



""" Some helpers for options
"""
import logging
from os.path import dirname, exists, join as pj
import pkg_resources

LOGGER = logging.getLogger(__name__)

available_options = {}
"""Dictionary of currently available option in pkglts. Discovered at run time"""

try:
    loc_input = raw_input
except NameError:
    loc_input = input


class Option(object):
    """Base class to store information associated with an option
    """

    def __init__(self):
        self._name = None
        self._root_dir = None
        self._update_parameters = None
        self._check = None
        self._require = None
        self._environment_extensions = None
        self._regenerate = None

    def from_entry_point(self, ept):
        """Associate an entry point with its correct attribute.

        Notes: does not load the entry point

        Args:
            ept (entry point): as returned by pkg_resources

        Returns:
            None
        """
        self._name, func_name = ept.name.split(".")
        if func_name == "root":
            self._root_dir = ept
        elif func_name == "update_parameters":
            self._update_parameters = ept
        elif func_name == "check":
            self._check = ept
        elif func_name == "require":
            self._require = ept
        elif func_name == "environment_extensions":
            self._environment_extensions = ept
        elif func_name == "regenerate":
            self._regenerate = ept
        else:
            # silently ignore other type of entry points
            LOGGER.error("unknown entry point attribute: '%s'", func_name)

    def root_dir(self):
        """Base directory containing option definition files."""
        if self._root_dir is None:
            raise UserWarning("Need to associate entry point first")

        if isinstance(self._root_dir, pkg_resources.EntryPoint):
            self._root_dir = dirname(self._root_dir.load().__file__)

        return self._root_dir

    def example_dir(self):
        """Directory containing option example files."""
        pth = pj(self.root_dir(), 'example')
        if exists(pth):
            return pth

        return None

    def resource_dir(self):
        """Directory containing option resource files."""
        pth = pj(self.root_dir(), 'resource')
        if exists(pth):
            return pth

        return None

    def update_parameters(self, cfg):
        """Update configuration with option parameters.

        Args:
            cfg (Config):  current package configuration

        Returns:
            (Config)
        """
        if self._update_parameters is None:
            cfg[self._name] = {}
            return

        if isinstance(self._update_parameters, pkg_resources.EntryPoint):
            self._update_parameters = self._update_parameters.load()

        return self._update_parameters(cfg)

    def check(self, *args, **kwds):
        """Check validity of parameters for this option.

        Returns:
            (list of str): list of failing params
        """
        if self._check is None:
            return []

        if isinstance(self._check, pkg_resources.EntryPoint):
            self._check = self._check.load()

        return self._check(*args, **kwds)

    def require(self, *args, **kwds):
        """Check dependencies for this option.

        Returns:
            (list of Dependency): list of modules or options this option require
        """
        if self._require is None:
            return []

        if isinstance(self._require, pkg_resources.EntryPoint):
            self._require = self._require.load()

        return self._require(*args, **kwds)

    def environment_extensions(self, *args, **kwds):
        """Get jinja2 environment extensions defined by this option.

        Returns:
            (list of func): extensions defined by this option
        """
        if self._environment_extensions is None:
            return {}

        if isinstance(self._environment_extensions, pkg_resources.EntryPoint):
            self._environment_extensions = self._environment_extensions.load()

        return self._environment_extensions(*args, **kwds)

    def regenerate(self, *args, **kwds):
        """Call regenerate associated with this option.

        Returns:
            (any)
        """
        if self._regenerate is None:
            return None

        if isinstance(self._regenerate, pkg_resources.EntryPoint):
            self._regenerate = self._regenerate.load()

        return self._regenerate(*args, **kwds)


def find_available_options():
    """Discover all available options.

    Returns:
        (list of Option)
    """
    for ept in pkg_resources.iter_entry_points(group='pkglts'):
        option_name = ept.name.split(".")[0]
        if option_name not in available_options:
            available_options[option_name] = Option()

        opt = available_options[option_name]
        opt.from_entry_point(ept)


def get_user_permission(action_name, default_true=True):
    """Helper function to input a yes or no question.

    Args:
        action_name (str): action to ask permission for.
        default_true (bool): whether answer is yes or no by default.

    Returns:
        (bool)
    """
    if default_true:
        ans = loc_input("%s [y], n?" % action_name) in ("", "y")
    else:
        ans = loc_input("%s y, [n]?" % action_name) == "y"

    return ans

#
# def get_key(key, env):
#     """ Fetch a specific key in env
#     """
#     try:
#         elms = key.split(".")
#         d = env
#         for k in elms:
#             d = d[k]
#
#         return d
#     except KeyError:
#         return None
#
#
# def _ask_single_arg(name, default):
#     # prompt user for value
#     msg = name
#     if default is None:
#         msg += ":"
#         val = ""
#     else:
#         val = str(default)
#         msg += " [%s]:" % val
#
#     ans = loc_input(msg)
#     if ans == "":
#         return val
#     else:
#         return ans
#
#
# def ask_arg(name, pkg_cfg=None, default=None, extra=None):
#     """ Prompt the user for the value of some argument
#
#     If user returns nothing then default will be returned
#
#     If pkg_cfg is provided, the function will attempt to
#     find a proper default in it.
#
#     If extra is provided, the function will attempt to
#     find a value in it and return it without prompting
#     the user.
#
#     If default is a list of elements, even empty, the function
#     will return a list of answers.
#
#     args:
#      - name (str): name of argument. Can be in the form of
#                    'option.arg'.
#      - pkg_cfg (dict of (str, any)): place to look for defaults
#      - default (any): default to use as a last resort
#      - extra (dict of (str, any)): place to look for values
#
#     return:
#      - (str): value of argument either from prompt or from default
#     """
#     if extra is None:
#         extra = {}
#
#     # try to find value in extra
#     if name in extra:
#         return extra[name]
#
#     key = name.split(".")[-1]
#     if key in extra:
#         return extra[key]
#
#     if pkg_cfg is None:
#         pkg_cfg = {}
#
#     if isinstance(default, (list, tuple)):
#         # ask multiple args
#         val = get_key(name, pkg_cfg)
#         if val is None:
#             val = default
#
#         if not isinstance(val, (list, tuple)):
#             msg = "bad value in pkg_cfg, expected list for %s" % name
#             raise UserWarning(msg)
#
#         i = 0
#         items = []
#         res = "txt"
#         while res != "":
#             if i < len(val):
#                 loc_val = val[i]
#             else:
#                 loc_val = ""
#             res = _ask_single_arg(name, loc_val)
#             if res != "":
#                 items.append(res)
#                 i += 1
#
#         return items
#     else:
#         val = get_key(name, pkg_cfg)
#         if val is None:
#             val = default
#         return _ask_single_arg(name, val)



"""Pip front end to install package in current environment.
"""
import imp
import pip
from pip import get_installed_distributions
from pip import main as pip_install


def installed_packages():
    """Iterate on all packages installed in the current python environment.

    return:
        (iter of str)
    """
    pip.utils.pkg_resources = imp.reload(pip.utils.pkg_resources)
    for pjt in get_installed_distributions():
        yield pjt.project_name


def install(name):
    """Install a package in the current python environment.

    arg:
     - name (str): name of the package

    return:
     - (bool): whether installation was successful or not
    """
    pip_install(['install', name])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# {# pkglts, pysetup.kwds
# format setup arguments

from os import walk
from os.path import abspath, normpath, splitext
from os.path import join as pj

from setuptools import setup, find_packages


short_descr = "Building packages with long term support"
readme = open('README.rst').read()
history = open('HISTORY.rst').read()


# find version number in src/pkglts/version.py
version = {}
with open("src/pkglts/version.py") as fp:
    exec(fp.read(), version)

# find packages
pkgs = find_packages('src')

pkg_data = {}

nb = len(normpath(abspath("src/pkglts"))) + 1
data_rel_pth = lambda pth: normpath(abspath(pth))[nb:]

data_files = []
for root, dnames, fnames in walk("src/pkglts"):
    for name in fnames:
        if splitext(name)[-1] in ['', '.bat', '.cfg', '.json', '.md', '.in', '.ini', '.png', '.ps1', '.rst', '.sh', '.tpl', '.txt', '.yaml', '.yml']:
            data_files.append(data_rel_pth(pj(root, name)))


pkg_data['pkglts'] = data_files

setup_kwds = dict(
    name='pkglts',
    version=version["__version__"],
    description=short_descr,
    long_description=readme + '\n\n' + history,
    author="revesansparole",
    author_email="revesansparole@gmail.com",
    url='https://github.com/revesansparole/pkglts',
    license='CeCILL-C',
    zip_safe=False,

    packages=pkgs,
    package_dir={'': 'src'},
    
    
    package_data=pkg_data,
    setup_requires=[
        "pytest-runner",
        ],
    install_requires=[
        "jinja2",
        "requests",
        "semver",
        "unidecode",
        ],
    tests_require=[
        "coverage",
        "coveralls",
        "flake8",
        "nbconvert",
        "pytest",
        "pytest-cov",
        "pytest-mock",
        "sphinx",
        "tox",
        "twine",
        ],
    entry_points={},
    keywords='packaging, package builder',
    
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    )
# #}
# change setup_kwds below before the next pkglts tag

setup_kwds['entry_points']['console_scripts'] = ['pmg = pkglts.manage_script:main']
setup_kwds['entry_points']['pkglts'] = [
    'appveyor.root = pkglts.option.appveyor',
    'appveyor.update_parameters = pkglts.option.appveyor.config:update_parameters',
    'appveyor.check = pkglts.option.appveyor.config:check',
    'appveyor.require = pkglts.option.appveyor.config:require',
    'appveyor.environment_extensions = pkglts.option.appveyor.handlers:environment_extensions',

    'base.root = pkglts.option.base',
    'base.update_parameters = pkglts.option.base.config:update_parameters',
    'base.check = pkglts.option.base.config:check',
    'base.require = pkglts.option.base.config:require',
    'base.environment_extensions = pkglts.option.base.handlers:environment_extensions',

    'conda.root = pkglts.option.conda',
    'conda.require = pkglts.option.conda.config:require',

    'coverage.root = pkglts.option.coverage',
    'coverage.require = pkglts.option.coverage.config:require',

    'coveralls.root = pkglts.option.coveralls',
    'coveralls.require = pkglts.option.coveralls.config:require',
    'coveralls.environment_extensions = pkglts.option.coveralls.handlers:environment_extensions',

    'data.root = pkglts.option.data',
    'data.update_parameters = pkglts.option.data.config:update_parameters',
    'data.check = pkglts.option.data.config:check',
    'data.require = pkglts.option.data.config:require',

    'doc.root = pkglts.option.doc',
    'doc.update_parameters = pkglts.option.doc.config:update_parameters',
    'doc.check = pkglts.option.doc.config:check',
    'doc.require = pkglts.option.doc.config:require',

    'flake8.root = pkglts.option.flake8',
    'flake8.require = pkglts.option.flake8.config:require',

    'git.root = pkglts.option.git',
    'git.update_parameters = pkglts.option.git.config:update_parameters',
    'git.check = pkglts.option.git.config:check',
    'git.require = pkglts.option.git.config:require',
    'git.environment_extensions = pkglts.option.git.handlers:environment_extensions',

    'github.root = pkglts.option.github',
    'github.update_parameters = pkglts.option.github.config:update_parameters',
    'github.check = pkglts.option.github.config:check',
    'github.require = pkglts.option.github.config:require',

    'gitlab.root = pkglts.option.gitlab',
    'gitlab.update_parameters = pkglts.option.gitlab.config:update_parameters',
    'gitlab.check = pkglts.option.gitlab.config:check',
    'gitlab.require = pkglts.option.gitlab.config:require',

    'landscape.root = pkglts.option.landscape',
    'landscape.require = pkglts.option.landscape.config:require',
    'landscape.environment_extensions = pkglts.option.landscape.handlers:environment_extensions',

    'license.root = pkglts.option.license',
    'license.update_parameters = pkglts.option.license.config:update_parameters',
    'license.check = pkglts.option.license.config:check',
    'license.require = pkglts.option.license.config:require',
    'license.environment_extensions = pkglts.option.license.handlers:environment_extensions',

    'notebook.root = pkglts.option.notebook',
    'notebook.update_parameters = pkglts.option.notebook.config:update_parameters',
    'notebook.check = pkglts.option.notebook.config:check',
    'notebook.require = pkglts.option.notebook.config:require',
    'notebook.regenerate = pkglts.option.notebook.regenerate:main',

    'plugin_project.root = pkglts.option.plugin_project',
    'plugin_project.update_parameters = pkglts.option.plugin_project.config:update_parameters',
    'plugin_project.check = pkglts.option.plugin_project.config:check',
    'plugin_project.require = pkglts.option.plugin_project.config:require',

    'pypi.root = pkglts.option.pypi',
    'pypi.update_parameters = pkglts.option.pypi.config:update_parameters',
    'pypi.check = pkglts.option.pypi.config:check',
    'pypi.require = pkglts.option.pypi.config:require',
    'pypi.environment_extensions = pkglts.option.pypi.handlers:environment_extensions',

    'pysetup.root = pkglts.option.pysetup',
    'pysetup.update_parameters = pkglts.option.pysetup.config:update_parameters',
    'pysetup.check = pkglts.option.pysetup.config:check',
    'pysetup.require = pkglts.option.pysetup.config:require',
    'pysetup.environment_extensions = pkglts.option.pysetup.handlers:environment_extensions',

    'readthedocs.root = pkglts.option.readthedocs',
    'readthedocs.update_parameters = pkglts.option.readthedocs.config:update_parameters',
    'readthedocs.check = pkglts.option.readthedocs.config:check',
    'readthedocs.require = pkglts.option.readthedocs.config:require',
    'readthedocs.environment_extensions = pkglts.option.readthedocs.handlers:environment_extensions',

    'requires.root = pkglts.option.requires',
    'requires.require = pkglts.option.requires.config:require',
    'requires.environment_extensions = pkglts.option.requires.handlers:environment_extensions',

    'sphinx.root = pkglts.option.sphinx',
    'sphinx.update_parameters = pkglts.option.sphinx.config:update_parameters',
    'sphinx.check = pkglts.option.sphinx.config:check',
    'sphinx.require = pkglts.option.sphinx.config:require',

    'test.root = pkglts.option.test',
    'test.update_parameters = pkglts.option.test.config:update_parameters',
    'test.check = pkglts.option.test.config:check',
    'test.require = pkglts.option.test.config:require',

    'tox.root = pkglts.option.tox',
    'tox.require = pkglts.option.tox.config:require',

    'travis.root = pkglts.option.travis',
    'travis.require = pkglts.option.travis.config:require',
    'travis.environment_extensions = pkglts.option.travis.handlers:environment_extensions',

    'version.root = pkglts.option.version',
    'version.update_parameters = pkglts.option.version.config:update_parameters',
    'version.check = pkglts.option.version.config:check',
    'version.require = pkglts.option.version.config:require',
]

# do not change things below
# {# pkglts, pysetup.call
setup(**setup_kwds)
# #}



"""
Set of function to extend jinja2 environment.
"""
import logging
import re
import subprocess

from unidecode import unidecode

LOGGER = logging.getLogger(__name__)


def environment_extensions(cfg):
    """Add more functionality to an environment.

    Args:
        cfg (Config):  current package configuration

    Returns:
        dict of str: any
    """
    del cfg

    try:
        log = subprocess.check_output(['git', 'log', '--all', '--use-mailmap']).decode('utf-8')
        commiters = re.findall(r'Author: (.* <.*@.*>)\n', unidecode(log))
        ccs = [(commiters.count(name), name) for name in set(commiters)]
        contributors = [name for nb, name in sorted(ccs, reverse=True)]
    except KeyError:
        LOGGER.warning("Please add git to your $PATH")
        contributors = ["I failed to construct the contributor list"]
    except subprocess.CalledProcessError as err:
        contributors = ["Pb with git, %s" % str(err)]

    return {'contributors': contributors}



# import github3


# def fetch_github_version(txt, env):
#     if txt == 'off':
#         return ""
#
#     # fetch version
#
#     # store it in env
#     cfg = env['version']
#     cfg['major'] = "0"
#     cfg['minor'] = "9"
#     cfg['post'] = "1"
#
#     # return no txt
#     return ""



"""
Set of function to associate hash to blobs.
"""
from base64 import b64encode
from hashlib import sha512
import json
import logging
from os.path import join as pj
from os.path import normpath

from .config import pkglts_dir, pkg_hash_file
from .templating import parse_source

LOGGER = logging.getLogger(__name__)


def pth_as_key(pth):
    """Normalize path to enable to use them as keys

    Args:
        pth (str): path to normalize

    Returns:
        (str)
    """
    return normpath(pth).replace("\\", "/")


def compute_hash(txt):
    """Compute hash summary of a text

    Args:
        txt (str): content to hash

    Returns:
        (str): hash key
    """
    algo = sha512()
    algo.update(txt.encode('utf-8'))  # TODO bad if non utf-8 encoded file
    return b64encode(algo.digest()).decode('utf-8')


def get_pkg_hash(rep="."):
    """Read pkg_hash file associated to this package

    Args:
        rep (str): directory to search for info

    Returns:
        (dict of str, list): hash map of preserved sections in this
                             package
    """
    with open(pj(rep, pkglts_dir, pkg_hash_file), 'r') as fhr:
        return json.load(fhr)


def write_pkg_hash(pkg_hash, rep="."):
    """Store hash associated to this package on disk.

    Args:
        pkg_hash (dict of str, list): hash map of preserved sections in this
                                     package
        rep (str): directory to search for info

    Returns:
        None
    """
    LOGGER.info("write package hash")
    cfg = dict(pkg_hash)

    with open(pj(rep, pkglts_dir, pkg_hash_file), 'w') as fhw:
        json.dump(cfg, fhw, sort_keys=True, indent=2)


def modified_file_hash(pth, pkg_hash):
    """Check whether a file complies with previously stored hash

    Args:
        pth (str): path to file to test
        pkg_hash (dict of str, list): hash map of preserved sections in this
                                     package

    Returns:
        (bool): whether this file has been modified or not
    """
    key = pth_as_key(pth)
    if key not in pkg_hash:
        raise IOError("%s not under check" % key)

    ref_blocks = pkg_hash[key]

    with open(pth, 'r') as fhr:
        blocks = parse_source(fhr.read())

    lts_blocks = dict((block.bid, block.content) for block in blocks
                      if block.bid is not None)
    if set(lts_blocks) != set(ref_blocks):
        return True

    for bid, cnt in lts_blocks.items():
        sha = compute_hash(cnt)
        if sha != ref_blocks[bid]:
            return True

    return False



"""
Set of function related to handling the configuration of this option.
"""
from pkglts.dependency import Dependency


def update_parameters(cfg):
    """Update config with parameters necessary for this option.

    Notes: create a section with option name to store params.

    Args:
        cfg (dict): dict of option parameters as seen in pkg_cfg.json

    Returns:
        None: update in place
    """
    sec = dict(
        theme="default",
        autodoc_dvlpt=True,
        build_dir="build/sphinx"
    )
    cfg['sphinx'] = sec


def check(cfg):
    """Check the validity of parameters in working environment.

    Args:
        cfg (Config):  current package configuration

    Returns:
        (list of str): list of faulty parameters
    """
    invalids = []
    if cfg['doc']['fmt'] != 'rst':
        invalids.append('doc.fmt')

    theme = cfg['sphinx']['theme']
    if theme != str(theme):
        invalids.append('sphinx.theme')

    return invalids


def require(purpose, cfg):
    """List of requirements for this option for a given purpose.

    Args:
        purpose (str): either 'option', 'setup', 'install' or 'dvlpt'
        cfg (Config):  current package configuration

    Returns:
        (list of Dependency)
    """
    if purpose == 'option':
        options = ['test', 'doc', 'license']
        return [Dependency(name) for name in options]

    if purpose == 'dvlpt':
        deps = [Dependency('sphinx')]
        if cfg["sphinx"]["theme"] == "sphinx_rtd_theme":
            deps.append(Dependency('sphinx_rtd_theme'))

        return deps

    return []



"""
Set of functions to extend jinja2.
"""
from os.path import exists
import re

OPENING_MARKER = "{" + "#"
CLOSING_MARKER = "#" + "}"

BLOCK_RE = re.compile(
    r"\{#[ ]pkglts,[ ](?P<key>[a-zA-Z0-9.]*)(?P<aft_head>.*?)\n(?P<cnt>.*?)#\}(?P<aft_foot>.*?(?=\n))",
    re.DOTALL | re.MULTILINE)


class TplBlock(object):
    """Simple container for templated blocks."""

    def __init__(self):
        self.before_header = ""
        self.bid = None
        self.after_header = ""
        self.content = ""
        self.before_footer = ""
        self.after_footer = ""

    def __str__(self):
        return "Block(\n|%s|%s|%s|\n|%s|\n|%s|%s|\n)" % (self.before_header,
                                                         self.bid,
                                                         self.after_header,
                                                         repr(self.content),
                                                         self.before_footer,
                                                         self.after_footer)

    def pkglts_defined(self, bid, cnt):
        """Create a block regenerated by pkglts.

        Args:
            bid (str): id for this block (unique per file only)
            cnt (str): content

        Returns:
            None
        """
        self.bid = bid
        self.content = cnt

    def user_defined(self, cnt):
        """Make this block a user defined one.

        Args:
            cnt (str): content

        Returns:
            None
        """
        self.bid = None
        self.content = cnt


def parse_source(txt):
    """Parse text to find preserved blocks

    Args:
        txt (str): full text to parse

    Returns:
        (list of [str, str, str, str]): ordered list of blocks:
                    - block_id (or None if content is not preserved
                    - line start before start for preserved content
                    - content
                    - line start before end for preserved content
    """
    blocks = []
    last_end = 0

    for res in BLOCK_RE.finditer(txt):
        i = res.start()
        while i > last_end and txt[i] != "\n":
            i -= 1

        if i >= last_end:
            block = TplBlock()
            block.user_defined(txt[last_end: (i + 1)])
            blocks.append(block)

        bef = txt[(i + 1): res.start()]

        bid = res.group('key')
        aft_head = res.group('aft_head')
        cnt = res.group('cnt')
        aft_foot = res.group('aft_foot')

        i = len(cnt) - 1
        while i > 0 and cnt[i] != "\n":
            i -= 1
        aft = cnt[(i + 1): len(cnt)]
        cnt = cnt[:i]
        last_end = res.end() + 1

        block = TplBlock()
        block.pkglts_defined(bid, cnt)
        block.before_header = bef
        block.after_header = aft_head
        block.before_footer = aft
        block.after_footer = aft_foot
        blocks.append(block)

    if last_end < len(txt):
        block = TplBlock()
        block.user_defined(txt[last_end: len(txt)])
        blocks.append(block)

    return blocks


def render(cfg, src_pth, tgt_pth):
    """Render src_pth templated file into tgt_pth

    Notes: keeps 'preserved' block structure

    Args:
        cfg (Config):  current package configuration
        src_pth (str): path to reference file
        tgt_pth (str): path to potentially non existent yet target file

    Returns:
        (list of [str, str]): key, cnt for preserved blocks
    """
    # parse src file to find 'preserved' blocks
    with open(src_pth, 'r') as fhr:
        src_blocks = parse_source(fhr.read())

    blocks = []
    if exists(tgt_pth):  # retrieves preserved blocks from source
        # parse tgt file to find 'preserved' blocks
        with open(tgt_pth, 'r') as fhr:
            tgt_blocks = parse_source(fhr.read())

        src_blocks = dict((b.bid, b) for b in src_blocks if b.bid is not None)
        for tgt_block in tgt_blocks:
            # check stored content hash
            if tgt_block.bid is not None:
                tgt_block.content = src_blocks[tgt_block.bid].content

            blocks.append(tgt_block)
    else:  # format non preserved blocks for the first and only time
        for block in src_blocks:
            if block.bid is None:
                block.content = cfg.render(block.content)

            blocks.append(block)

    # regenerate preserved block content
    preserved = []
    tgt = ""
    for block in blocks:
        if block.bid is None:
            tgt += block.content
        else:
            # format cnt
            cnt = cfg.render(block.content)
            preserved.append((block.bid, cnt))
            # rewrite preserved tag if necessary
            tgt += block.before_header + "{" + "# pkglts, %s" % block.bid + "%s\n" % block.after_header
            tgt += cnt
            tgt += "\n" + block.before_footer + "#" + "}%s\n" % block.after_footer

    with open(tgt_pth, 'w') as fhw:
        fhw.write(tgt)

    return preserved



"""Simple front end to emulate all environment modification
by printing the required operations.
"""
from os import listdir
from os.path import basename, splitext
import sys


def installed_packages():
    """Iterate on all packages installed in the current python environment.

    .. warning: extremely simple implementation that list the content
                of site-packages only.

    return:
        (list of str)
    """
    for pth in sys.path:
        name = basename(pth)
        if name == "site-packages":
            for pkg in listdir(pth):
                pkgname, ext = splitext(pkg)
                if ext in ("", ".py", ".egg", ".egg-link"):
                    if "-" in pkgname:
                        pkgname = pkgname.split("-")[0]

                    yield pkgname


def install(name):
    """Install a package in the current python environment.

    arg:
     - name (str): name of the package

    return:
     - (bool): whether installation was successful or not
    """
    print("install %s" % name)



from pkglts.config_management import Config
from pkglts.option.doc.config import check, require, update_parameters


def test_update_parameters():
    cfg = {}
    update_parameters(cfg)
    assert len(cfg['doc']) == 3


def test_config_check_description_exists():
    cfg = Config(dict(doc={'description': "mydescr", 'fmt': 'rst', 'keywords': []}))
    assert cfg['doc']['description'] == "mydescr"
    assert 'doc.description' not in check(cfg)


def test_config_check_description_valid():
    cfg = Config(dict(doc={'description': "", 'fmt': 'rst', 'keywords': []}))
    assert 'doc.description' in check(cfg)


def test_config_check_keywords_exists():
    cfg = Config(dict(doc={'description': "mydescr", 'fmt': 'rst', 'keywords': []}))
    assert len(cfg['doc']['keywords']) == 0
    assert 'doc.keywords' not in check(cfg)


def test_config_check_fmt_valid():
    cfg = Config(dict(doc={'description': "mydescr", 'fmt': 'walou', 'keywords': []}))
    assert 'doc.fmt' in check(cfg)


def test_config_check_fmt_valid():
    cfg = Config(dict(doc={'description': "mydescr", 'fmt': 'walou', 'keywords': []}))
    assert 'doc.fmt' in check(cfg)


def test_require():
    cfg = Config(dict(base={}, doc={'fmt': 'rst'}))
    
    assert len(require('option', cfg)) == 1
    assert len(require('setup', cfg)) == 0
    assert len(require('install', cfg)) == 0
    assert len(require('dvlpt', cfg)) == 0

    cfg = Config(dict(base={}, doc={'fmt': 'md'}))
    assert len(require('dvlpt', cfg)) == 1
